# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import FactorizedTensor from tensorly for tensorized operations
import tensorly as tl
import torch
import torch.nn as nn
import torch.nn.functional as F

tl.set_backend("pytorch")
import torch_harmonics as th
import torch_harmonics.distributed as thd

# from tensorly.plugins import use_opt_einsum
# use_opt_einsum('optimal')
from tltorch.factorized_tensors.core import FactorizedTensor

# import convenience functions for factorized tensors
from .activations import ComplexReLU

# for the experimental module
from .contractions import (
    _contract_localconv_fwd,
    compl_exp_mul2d_fwd,
    compl_exp_muladd2d_fwd,
    compl_mul2d_fwd,
    compl_muladd2d_fwd,
    real_mul2d_fwd,
    real_muladd2d_fwd,
)
from .factorizations import get_contract_fun


class SpectralConvS2(nn.Module):
    """
    Spectral Convolution according to Driscoll & Healy. Designed for convolutions on
    the two-sphere S2 using the Spherical Harmonic Transforms in torch-harmonics, but
    supports convolutions on the periodic domain via the RealFFT2 and InverseRealFFT2
    wrappers.
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        in_channels,
        out_channels,
        scale="auto",
        operator_type="diagonal",
        rank=0.2,
        factorization=None,
        separable=False,
        decomposition_kwargs=dict(),
        bias=False,
        use_tensorly=True,
        filter_residual: bool = False,
    ):  # pragma: no cover
        super(SpectralConvS2, self).__init__()

        if scale == "auto":
            scale = 1 / (in_channels * out_channels)

        self.forward_transform = forward_transform
        self.inverse_transform = inverse_transform

        self.modes_lat = self.inverse_transform.lmax
        self.modes_lon = self.inverse_transform.mmax

        self._round_trip_residual = filter_residual or (
            (self.forward_transform.nlat != self.inverse_transform.nlat)
            or (self.forward_transform.nlon != self.inverse_transform.nlon)
            or (self.forward_transform.grid != self.inverse_transform.grid)
        )
        # Make sure we are using a Complex Factorized Tensor
        if factorization is None:
            factorization = "Dense"  # No factorization

        if not factorization.lower().startswith("complex"):
            factorization = f"Complex{factorization}"

        # remember factorization details
        self.operator_type = operator_type
        self.rank = rank
        self.factorization = factorization
        self.separable = separable

        assert self.inverse_transform.lmax == self.modes_lat
        assert self.inverse_transform.mmax == self.modes_lon

        weight_shape = [in_channels]

        if not self.separable:
            weight_shape += [out_channels]

        if isinstance(self.inverse_transform, thd.DistributedInverseRealSHT):
            self.modes_lat_local = self.inverse_transform.lmax_local
            self.modes_lon_local = self.inverse_transform.mmax_local
            self.lpad_local = self.inverse_transform.lpad_local
            self.mpad_local = self.inverse_transform.mpad_local
        else:
            self.modes_lat_local = self.modes_lat
            self.modes_lon_local = self.modes_lon
            self.lpad = 0
            self.mpad = 0

        # padded weights
        # if self.operator_type == 'diagonal':
        #     weight_shape += [self.modes_lat_local+self.lpad_local, self.modes_lon_local+self.mpad_local]
        # elif self.operator_type == 'dhconv':
        #     weight_shape += [self.modes_lat_local+self.lpad_local]
        # else:
        #     raise ValueError(f"Unsupported operator type f{self.operator_type}")

        # unpadded weights
        if self.operator_type == "diagonal":
            weight_shape += [self.modes_lat_local, self.modes_lon_local]
        elif self.operator_type == "dhconv":
            weight_shape += [self.modes_lat_local]
        else:
            raise ValueError(f"Unsupported operator type f{self.operator_type}")

        if use_tensorly:
            # form weight tensors
            self.weight = FactorizedTensor.new(
                weight_shape,
                rank=self.rank,
                factorization=factorization,
                fixed_rank_modes=False,
                **decomposition_kwargs,
            )
            # initialization of weights
            self.weight.normal_(0, scale)
        else:
            assert factorization == "ComplexDense"
            self.weight = nn.Parameter(scale * torch.randn(*weight_shape, 2))
            if self.operator_type == "dhconv":
                self.weight.is_shared_mp = ["matmul", "w"]
            else:
                self.weight.is_shared_mp = ["matmul"]

        # get the contraction handle
        self._contract = get_contract_fun(
            self.weight, implementation="factorized", separable=separable
        )

        if bias:
            self.bias = nn.Parameter(scale * torch.zeros(1, out_channels, 1, 1))

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        residual = x
        x = x.float()
        B, C, H, W = x.shape

        with torch.amp.autocast("cuda", enabled=False):
            x = self.forward_transform(x)
            if self._round_trip_residual:
                x = x.contiguous()
                residual = self.inverse_transform(x)
                residual = residual.to(dtype)

        # approach with unpadded weights
        xp = torch.zeros_like(x)
        xp[..., : self.modes_lat_local, : self.modes_lon_local] = self._contract(
            x[..., : self.modes_lat_local, : self.modes_lon_local],
            self.weight,
            separable=self.separable,
            operator_type=self.operator_type,
        )
        x = xp.contiguous()

        # # approach with padded weights
        # x = self._contract(x, self.weight, separable=self.separable, operator_type=self.operator_type)
        # x = x.contiguous()

        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        if hasattr(self, "bias"):
            x = x + self.bias

        x = x.type(dtype)

        return x, residual


class LocalConvS2(nn.Module):
    """
    S2 Convolution according to Driscoll & Healy
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        in_channels,
        out_channels,
        nradius=120,
        scale="auto",
        bias=False,
    ):  # pragma: no cover
        super(LocalConvS2, self).__init__()

        if scale == "auto":
            scale = 1 / (in_channels * out_channels)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.nradius = nradius

        self.forward_transform = forward_transform
        self.zonal_transform = th.RealSHT(
            forward_transform.nlat,
            1,
            lmax=forward_transform.lmax,
            mmax=1,
            grid=forward_transform.grid,
        ).float()
        self.inverse_transform = inverse_transform

        self.modes_lat = self.inverse_transform.lmax
        self.modes_lon = self.inverse_transform.mmax
        self.output_dims = (self.inverse_transform.nlat, self.inverse_transform.nlon)

        assert self.inverse_transform.lmax == self.modes_lat
        assert self.inverse_transform.mmax == self.modes_lon

        self.weight = nn.Parameter(
            scale * torch.randn(in_channels, out_channels, nradius, 1)
        )

        self._contract = _contract_localconv_fwd

        if bias:
            self.bias = nn.Parameter(
                scale * torch.randn(1, out_channels, *self.output_dims)
            )

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        x = x.float()
        B, C, H, W = x.shape

        with torch.amp.autocast("cuda", enabled=False):
            f = torch.zeros(
                (self.in_channels, self.out_channels, H, 1),
                dtype=x.dtype,
                device=x.device,
            )
            f[..., : self.nradius, :] = self.weight

            x = self.forward_transform(x)
            f = self.zonal_transform(f)[..., :, 0]

            x = torch.view_as_real(x)
            f = torch.view_as_real(f)

        x = self._contract(x, f)
        x = x.contiguous()

        x = torch.view_as_complex(x)

        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        if hasattr(self, "bias"):
            x = x + self.bias

        x = x.type(dtype)

        return x


class SpectralAttentionS2(nn.Module):
    """
    Spherical non-linear FNO layer
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        operator_type="diagonal",
        sparsity_threshold=0.0,
        hidden_size_factor=2,
        complex_activation="real",
        scale="auto",
        bias=False,
        spectral_layers=1,
        drop_rate=0.0,
    ):  # pragma: no cover
        super(SpectralAttentionS2, self).__init__()

        self.embed_dim = embed_dim
        self.sparsity_threshold = sparsity_threshold
        self.operator_type = operator_type
        self.spectral_layers = spectral_layers

        if scale == "auto":
            self.scale = 1 / (embed_dim * embed_dim)

        self.modes_lat = forward_transform.lmax
        self.modes_lon = forward_transform.mmax

        # only storing the forward handle to be able to call it
        self.forward_transform = forward_transform
        self.inverse_transform = inverse_transform

        self.scale_residual = (
            self.forward_transform.nlat != self.inverse_transform.nlat
        ) or (self.forward_transform.nlon != self.inverse_transform.nlon)

        assert inverse_transform.lmax == self.modes_lat
        assert inverse_transform.mmax == self.modes_lon

        hidden_size = int(hidden_size_factor * self.embed_dim)

        if operator_type == "diagonal":
            self.mul_add_handle = compl_muladd2d_fwd
            self.mul_handle = compl_mul2d_fwd

            # weights
            w = [self.scale * torch.randn(self.embed_dim, hidden_size, 2)]
            for l in range(1, self.spectral_layers):
                w.append(self.scale * torch.randn(hidden_size, hidden_size, 2))
            self.w = nn.ParameterList(w)

            self.wout = nn.Parameter(
                self.scale * torch.randn(hidden_size, self.embed_dim, 2)
            )

            if bias:
                self.b = nn.ParameterList(
                    [
                        self.scale * torch.randn(hidden_size, 1, 1, 2)
                        for _ in range(self.spectral_layers)
                    ]
                )

            self.activations = nn.ModuleList([])
            for l in range(0, self.spectral_layers):
                self.activations.append(
                    ComplexReLU(
                        mode=complex_activation,
                        bias_shape=(hidden_size, 1, 1),
                        scale=self.scale,
                    )
                )

        elif operator_type == "l-dependant":
            self.mul_add_handle = compl_exp_muladd2d_fwd
            self.mul_handle = compl_exp_mul2d_fwd

            # weights
            w = [
                self.scale * torch.randn(self.modes_lat, self.embed_dim, hidden_size, 2)
            ]
            for l in range(1, self.spectral_layers):
                w.append(
                    self.scale
                    * torch.randn(self.modes_lat, hidden_size, hidden_size, 2)
                )
            self.w = nn.ParameterList(w)

            if bias:
                self.b = nn.ParameterList(
                    [
                        self.scale * torch.randn(hidden_size, 1, 1, 2)
                        for _ in range(self.spectral_layers)
                    ]
                )

            self.wout = nn.Parameter(
                self.scale * torch.randn(self.modes_lat, hidden_size, self.embed_dim, 2)
            )

            self.activations = nn.ModuleList([])
            for l in range(0, self.spectral_layers):
                self.activations.append(
                    ComplexReLU(
                        mode=complex_activation,
                        bias_shape=(hidden_size, 1, 1),
                        scale=self.scale,
                    )
                )

        else:
            raise ValueError("Unknown operator type")

        self.drop = nn.Dropout(drop_rate) if drop_rate > 0.0 else nn.Identity()

    def forward_mlp(self, x):  # pragma: no cover
        """forward pass of the MLP"""
        B, C, H, W = x.shape

        if self.operator_type == "block-separable":
            x = x.permute(0, 3, 1, 2)

        xr = torch.view_as_real(x)

        for l in range(self.spectral_layers):
            if hasattr(self, "b"):
                xr = self.mul_add_handle(xr, self.w[l], self.b[l])
            else:
                xr = self.mul_handle(xr, self.w[l])
            xr = torch.view_as_complex(xr)
            xr = self.activations[l](xr)
            xr = self.drop(xr)
            xr = torch.view_as_real(xr)

        # final MLP
        x = self.mul_handle(xr, self.wout)

        x = torch.view_as_complex(x)

        if self.operator_type == "block-separable":
            x = x.permute(0, 2, 3, 1)

        return x

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        residual = x
        x = x.to(torch.float32)

        # FWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = self.forward_transform(x)
            if self.scale_residual:
                x = x.contiguous()
                residual = self.inverse_transform(x)
                residual = residual.to(dtype)

        # MLP
        x = self.forward_mlp(x)

        # BWD transform
        x = x.contiguous()
        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        # cast back to initial precision
        x = x.to(dtype)

        return x, residual


class RealSpectralAttentionS2(nn.Module):
    """
    Non-linear SFNO layer using a real-valued NN instead of a complex one
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        operator_type="diagonal",
        sparsity_threshold=0.0,
        hidden_size_factor=2,
        complex_activation="real",
        scale="auto",
        bias=False,
        spectral_layers=1,
        drop_rate=0.0,
    ):  # pragma: no cover
        super(RealSpectralAttentionS2, self).__init__()

        self.embed_dim = embed_dim
        self.sparsity_threshold = sparsity_threshold
        self.operator_type = operator_type
        self.spectral_layers = spectral_layers

        if scale == "auto":
            self.scale = 1 / (embed_dim * embed_dim)

        self.modes_lat = forward_transform.lmax
        self.modes_lon = forward_transform.mmax

        # only storing the forward handle to be able to call it
        self.forward_transform = forward_transform
        self.inverse_transform = inverse_transform

        self.scale_residual = (
            self.forward_transform.nlat != self.inverse_transform.nlat
        ) or (self.forward_transform.nlon != self.inverse_transform.nlon)

        assert inverse_transform.lmax == self.modes_lat
        assert inverse_transform.mmax == self.modes_lon

        hidden_size = int(hidden_size_factor * self.embed_dim * 2)

        self.mul_add_handle = real_muladd2d_fwd
        self.mul_handle = real_mul2d_fwd

        # weights
        w = [self.scale * torch.randn(2 * self.embed_dim, hidden_size)]
        for l in range(1, self.spectral_layers):
            w.append(self.scale * torch.randn(hidden_size, hidden_size))
        self.w = nn.ParameterList(w)

        self.wout = nn.Parameter(
            self.scale * torch.randn(hidden_size, 2 * self.embed_dim)
        )

        if bias:
            self.b = nn.ParameterList(
                [
                    self.scale * torch.randn(hidden_size, 1, 1)
                    for _ in range(self.spectral_layers)
                ]
            )

        self.activations = nn.ModuleList([])
        for l in range(0, self.spectral_layers):
            self.activations.append(nn.ReLU())

        self.drop = nn.Dropout(drop_rate) if drop_rate > 0.0 else nn.Identity()

    def forward_mlp(self, x):  # pragma: no cover
        """forward pass of the MLP"""
        B, C, H, W = x.shape

        xr = torch.view_as_real(x)
        xr = xr.permute(0, 1, 4, 2, 3).reshape(B, C * 2, H, W)

        for l in range(self.spectral_layers):
            if hasattr(self, "b"):
                xr = self.mul_add_handle(xr, self.w[l], self.b[l])
            else:
                xr = self.mul_handle(xr, self.w[l])
            xr = self.activations[l](xr)
            xr = self.drop(xr)

        # final MLP
        xr = self.mul_handle(xr, self.wout)

        xr = xr.reshape(B, C, 2, H, W).permute(0, 1, 3, 4, 2)

        x = torch.view_as_complex(xr)

        return x

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        x = x.to(torch.float32)

        # FWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = self.forward_transform(x)

        # MLP
        x = self.forward_mlp(x)

        # BWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        # cast back to initial precision
        x = x.to(dtype)

        return x



import datetime
from dataclasses import asdict, dataclass, field
from typing import Any

from fme.core.training_history import get_job_env


@dataclass
class DatasetMetadata:
    source: dict[str, Any] = field(default_factory=dict)
    history: dict[str, Any] = field(default_factory=dict)
    title: str | None = None

    @classmethod
    def from_env(cls) -> "DatasetMetadata":
        """
        Create a DatasetMetadata instance from the job environment.
        """
        return cls(
            source=get_job_env(),
            history={"created": datetime.datetime.now().isoformat()},
        )

    def as_flat_str_dict(self) -> dict[str, str]:
        """
        Convert the metadata to a flat dictionary with string values. Nested
        dictionaries are flattened with dot notation for keys, e.g. {'source': {'b': 1}}
        becomes {'source.b': '1'}.
        """
        return asdict(self, dict_factory=_to_flat_str_dict)


def _to_flat_str_dict(fields: list[tuple[str, str | dict[str, Any]]]) -> dict[str, str]:
    """
    Convert a maybe-nested dictionary to a flat dictionary with string values.
    """
    flat_dict = {}
    for field_name, value in fields:
        if isinstance(value, dict):
            nested = _to_flat_str_dict([(k, v) for k, v in value.items()])
            flat_dict.update({f"{field_name}.{k}": v for k, v in nested.items()})
        else:
            flat_dict[field_name] = str(value)
    return flat_dict



import pytest
import torch

from fme.core.device import get_device

from .layers import ConditionalLayerNorm, Context, ContextConfig


@pytest.mark.parametrize("global_layer_norm", [True, False])
@pytest.mark.parametrize("n_channels", [32])
@pytest.mark.parametrize("embed_dim_scalar", [20, 0])
@pytest.mark.parametrize("embed_dim_2d", [20, 0])
@pytest.mark.parametrize("img_shape", [(8, 16)])
def test_conditional_layer_norm(
    n_channels: int,
    img_shape: tuple[int, int],
    global_layer_norm: bool,
    embed_dim_scalar: int,
    embed_dim_2d: int,
):
    epsilon = 1e-6
    device = get_device()
    conditional_layer_norm = ConditionalLayerNorm(
        n_channels,
        img_shape,
        context_config=ContextConfig(
            embed_dim_scalar=embed_dim_scalar,
            embed_dim_2d=embed_dim_2d,
        ),
        global_layer_norm=global_layer_norm,
        epsilon=epsilon,
    ).to(device)
    x = torch.randn(1, n_channels, *img_shape, device=device) * 5 + 2
    context_embedding = torch.randn(1, embed_dim_scalar, device=device)
    context_embedding_2d = torch.randn(1, embed_dim_2d, *img_shape, device=device)
    context = Context(context_embedding, context_embedding_2d)
    output = conditional_layer_norm(x, context)
    assert output.shape == x.shape
    torch.testing.assert_close(
        output.mean(), torch.tensor(0.0, device=device), atol=1e-3, rtol=0
    )
    torch.testing.assert_close(
        output.std(), torch.tensor(1.0, device=device), atol=1e-3, rtol=0
    )
    if not global_layer_norm:
        zero = torch.zeros(1, *img_shape, device=device)
        torch.testing.assert_close(output.mean(dim=1), zero, atol=1e-3, rtol=0)
        torch.testing.assert_close(
            (((n_channels - 1) / n_channels) ** 0.5 * output.std(dim=1) - 1),
            zero,
            atol=1e-3,
            rtol=0,
        )



from datetime import timedelta

import numpy as np
import torch
import xarray as xr

import fme
from fme.ace.data_loading.batch_data import BatchData
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.loss import WeightedMappingLossConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.optimization import NullOptimization
from fme.core.registry.module import ModuleSelector
from fme.core.timing import GlobalTimer

from ..ace.stepper import SingleModuleStepperConfig
from .multi_call import MultiCallConfig, get_multi_call_name

TEST_CONFIG = MultiCallConfig(
    forcing_name="CO2",
    forcing_multipliers={"_quadrupled_co2": 4, "_halved_co2": 0.5},
    output_names=["OLR", "ASR", "bar_0", "bar_10"],
)


def _step(input, next_step_forcing, *_):
    names = TEST_CONFIG.output_names
    prediction = {k: input["CO2"].detach().clone() for k in names}
    return prediction


def test_multi_call_names():
    assert set(TEST_CONFIG.names) == {
        "OLR_quadrupled_co2",
        "ASR_quadrupled_co2",
        "bar_quadrupled_co2_0",
        "bar_quadrupled_co2_10",
        "OLR_halved_co2",
        "ASR_halved_co2",
        "bar_halved_co2_0",
        "bar_halved_co2_10",
    }


def test_multi_call():
    config = TEST_CONFIG
    multi_call = config.build(_step)
    co2_value = 2.0
    shape = (1, 2, 3, 3)
    initial_condition = {"temperature": torch.ones(shape)}
    co2_data = {"CO2": torch.full(shape, co2_value)}

    output = multi_call.step(initial_condition | co2_data, {})

    assert set(output) == set(config.names)
    for name in config.output_names:
        for multiplier_name, multiplier_value in config.forcing_multipliers.items():
            # the _step method in this test module returns predictions equal
            # to the CO2 value in the forcing data, hence this check.
            multi_call_name = get_multi_call_name(name, multiplier_name)
            torch.testing.assert_close(
                output[multi_call_name],
                co2_data["CO2"] * multiplier_value,
            )
    torch.testing.assert_close(co2_data["CO2"], torch.full(shape, co2_value))


def get_scalar_data(names, value):
    return {n: value for n in names}


def _get_stepper_config(
    in_names, out_names, all_names, multi_call_config, include_loss
):
    class CustomModule(torch.nn.Module):
        def forward(self, x):
            channel_dim = -3
            input_co2 = x.select(channel_dim, 0)
            input_temperature = x.select(channel_dim, 1)
            new_temperature = input_temperature + input_co2
            output_olr = input_co2 * input_temperature
            return torch.stack([new_temperature, output_olr], dim=channel_dim)

        def eval(self):
            pass

    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": CustomModule()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means=get_scalar_data(all_names, 0.0),
            stds=get_scalar_data(all_names, 1.0),
        ),
        loss=WeightedMappingLossConfig(type="MSE", weights={"temperature": 1.0}),
        multi_call=multi_call_config,
        include_multi_call_in_loss=include_loss,
    )

    return config


def test_integration_with_stepper():
    img_shape = (5, 5)
    full_shape = (1, 3, *img_shape)
    horizontal_coord = LatLonCoordinates(
        torch.zeros(img_shape[0], device=fme.get_device()),
        torch.zeros(img_shape, device=fme.get_device()),
    )
    vertical_coord = HybridSigmaPressureCoordinate(
        torch.tensor([1.0], device=fme.get_device()),
        torch.tensor([0.0], device=fme.get_device()),
    )
    timestep = timedelta(seconds=1)

    in_names = ["CO2", "temperature"]
    out_names = ["temperature", "OLR"]
    multi_call_names = ["OLR_doubled_co2"]
    multi_call_config = MultiCallConfig(
        forcing_name="CO2",
        forcing_multipliers={"_doubled_co2": 2},
        output_names=["OLR"],
    )
    expected_all_names = set(in_names + out_names + multi_call_names)
    config = _get_stepper_config(
        in_names, out_names, expected_all_names, multi_call_config, True
    )

    assert set(config.diagnostic_names) == set(out_names).difference(in_names).union(
        multi_call_names
    )
    assert set(config.all_names) == expected_all_names
    stepper = config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coord,
            vertical_coordinate=vertical_coord,
            timestep=timestep,
        ),
    )
    time = xr.DataArray([[1, 1, 1]], dims=["sample", "time"])
    data = BatchData(
        {
            n: torch.ones(full_shape, device=fme.get_device())
            for n in expected_all_names
        },
        time,
    )
    with GlobalTimer():
        output = stepper.train_on_batch(data, NullOptimization())
    assert "OLR_doubled_co2" in output.gen_data
    assert "temperature" in output.gen_data
    assert "OLR" in output.gen_data
    assert "OLR_doubled_co2" in output.target_data
    expected_loss = output.metrics["loss_step_0"] + output.metrics["loss_step_1"]
    torch.testing.assert_close(output.metrics["loss"], expected_loss)
    # this value check is based on the implementation of the CustomModule above
    # first output time step
    co2 = data.data["CO2"][0, 0].cpu().numpy()  # assuming constant in time
    for t in range(3):
        output_temperature = output.gen_data["temperature"][0, 0, t].cpu().numpy()
        np.testing.assert_allclose(output_temperature, co2 + t)
        if t > 0:
            # only check diagnostic OLR for the output time steps, not initial condition
            input_temperature = (
                output.gen_data["temperature"][0, 0, t - 1].cpu().numpy()
            )
            olr = output.gen_data["OLR"][0, 0, t].cpu().numpy()
            np.testing.assert_allclose(olr, input_temperature * co2)
            olr_doubled = output.gen_data["OLR_doubled_co2"][0, 0, t].cpu().numpy()
            # note for the double-called case, it should use the temperature from the
            # non-doubled CO2 case
            np.testing.assert_allclose(olr_doubled, input_temperature * 2 * co2)

    # rerun without including multi-call in loss
    multi_call_config = MultiCallConfig(
        forcing_name="CO2",
        forcing_multipliers={"_doubled_co2": 2},
        output_names=["OLR"],
    )
    config = _get_stepper_config(
        in_names, out_names, expected_all_names, multi_call_config, False
    )
    stepper = config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coord,
            vertical_coordinate=vertical_coord,
            timestep=timestep,
        ),
    )
    with GlobalTimer():
        output_without_loss = stepper.train_on_batch(data, NullOptimization())

    torch.testing.assert_close(
        output_without_loss.metrics["loss"],
        output_without_loss.metrics["loss_step_0"]
        + output_without_loss.metrics["loss_step_1"],
    )



import random

import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.data_loading.augmentation import RotateModifier
from fme.ace.data_loading.batch_data import BatchData


def rotate(data: torch.Tensor) -> torch.Tensor:
    return torch.flip(data, dims=[-2, -1])


def test_rotate_modifier_all_rotation():
    rotate_modifier = RotateModifier(
        rotate_probability=1.0, additional_directional_names=[]
    )
    n_lat = 8
    n_lon = 16
    batch = BatchData(
        data={
            "UGRD": torch.randn(1, 2, n_lat, n_lon),
            "VGRD": torch.randn(1, 2, n_lat, n_lon),
            "PS": torch.randn(1, 2, n_lat, n_lon),
        },
        time=xr.DataArray(np.zeros((1, 2)), dims=["sample", "time"]),
        horizontal_dims=["lat", "lon"],
    )
    rotated_batch = rotate_modifier(batch)
    assert rotated_batch.data["UGRD"].shape == (1, 2, n_lat, n_lon)
    assert torch.allclose(rotate(rotated_batch.data["UGRD"]), -1 * batch.data["UGRD"])
    assert torch.allclose(rotate(rotated_batch.data["VGRD"]), -1 * batch.data["VGRD"])
    assert torch.allclose(rotate(rotated_batch.data["PS"]), batch.data["PS"])


def test_rotate_modifier_no_rotation():
    rotate_modifier = RotateModifier(
        rotate_probability=0.0, additional_directional_names=[]
    )
    n_lat = 8
    n_lon = 16
    batch = BatchData(
        data={
            "UGRD": torch.randn(1, 2, n_lat, n_lon),
            "VGRD": torch.randn(1, 2, n_lat, n_lon),
            "PS": torch.randn(1, 2, n_lat, n_lon),
        },
        time=xr.DataArray(np.zeros((1, 2)), dims=["sample", "time"]),
        horizontal_dims=["lat", "lon"],
    )
    rotated_batch = rotate_modifier(batch)
    assert rotated_batch.data["UGRD"].shape == (1, 2, n_lat, n_lon)
    assert torch.allclose(rotated_batch.data["UGRD"], batch.data["UGRD"])
    assert torch.allclose(rotated_batch.data["VGRD"], batch.data["VGRD"])
    assert torch.allclose(rotated_batch.data["PS"], batch.data["PS"])


def test_rotate_modifier_random_rotation():
    random.seed(0)
    rotate_modifier = RotateModifier(
        rotate_probability=0.5, additional_directional_names=[]
    )
    n_lat = 8
    n_lon = 16
    batch = BatchData(
        data={
            "UGRD": torch.randn(40, 2, n_lat, n_lon),
            "VGRD": torch.randn(40, 2, n_lat, n_lon),
            "PS": torch.randn(40, 2, n_lat, n_lon),
        },
        time=xr.DataArray(np.zeros((40, 2)), dims=["sample", "time"]),
        horizontal_dims=["lat", "lon"],
    )
    rotated_batch = rotate_modifier(batch)
    assert rotated_batch.data.keys() == batch.data.keys()
    assert rotated_batch.data["UGRD"].shape == (40, 2, n_lat, n_lon)
    rotated = {}
    unrotated = {}
    for name in rotated_batch.data:
        unrotated[name] = np.all(
            torch.abs(batch.data[name] - rotated_batch.data[name]).cpu().numpy() < 1e-6,
            axis=(1, 2, 3),
        )
        if name in ("UGRD", "VGRD"):
            sign = -1
        else:
            sign = 1
        rotated[name] = np.all(
            torch.abs(sign * rotate(batch.data[name]) - rotated_batch.data[name])
            .cpu()
            .numpy()
            < 1e-6,
            axis=(1, 2, 3),
        )
        assert np.all(rotated[name] + unrotated[name] == 1), name
        assert np.sum(rotated[name]) > 0, name
        assert np.sum(unrotated[name]) > 0, name
    for name in ("VGRD", "PS"):
        assert np.all(rotated[name] == rotated["UGRD"]), name
        assert np.all(unrotated[name] == unrotated["UGRD"]), name


@pytest.mark.parametrize(
    "name, additional_directional_names, match_expected",
    [
        ("UGRD", [], True),
        ("VGRD", [], True),
        ("UGRD_10m", [], True),
        ("UGRD_10m", ["UGRD"], True),
        ("VGRD200", [], True),
        ("eastward_wind_3", [], True),
        ("UGRD10m", [], True),
        ("NWIND10m", [], False),
        ("NWIND10m", ["NWIND"], True),
    ],
)
def test_rotate_modifier_pattern(
    name: str, additional_directional_names: list[str], match_expected: bool
):
    rotate_modifier = RotateModifier(
        rotate_probability=1.0,
        additional_directional_names=additional_directional_names,
    )
    assert (rotate_modifier._pattern.match(name) is not None) == match_expected, name



from collections.abc import Sequence
from typing import Self

import torch
import xarray as xr

from fme.core.dataset.subset import XarraySubset
from fme.core.dataset.xarray import XarrayDataset
from fme.core.typing_ import TensorDict


class XarrayConcat(torch.utils.data.Dataset):
    def __init__(self, datasets: Sequence[XarrayDataset | XarraySubset | Self]):
        self._dataset = torch.utils.data.ConcatDataset(datasets)
        sample_start_times = datasets[0].sample_start_times
        for dataset in datasets[1:]:
            sample_start_times = sample_start_times.append(dataset.sample_start_times)
            assert dataset.sample_n_times == datasets[0].sample_n_times
            if not dataset.sample_n_times == datasets[0].sample_n_times:
                raise ValueError(
                    "All concatenated datasets \
                         must have the same number of steps per sample item."
                )
        self._sample_start_times = sample_start_times
        assert len(self._dataset) == len(sample_start_times)
        self._sample_n_times = datasets[0].sample_n_times

        for dataset in datasets[1:]:
            if dataset.dims != datasets[0].dims:
                raise ValueError(
                    "Datasets being concatenated do not have the same dimensions: "
                    f"{dataset.dims} != {datasets[0].dims}"
                )
        self.dims: list[str] = datasets[0].dims

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        return self._dataset[idx]

    @property
    def sample_start_times(self):
        return self._sample_start_times

    @property
    def sample_n_times(self) -> int:
        """The length of the time dimension of each sample."""
        return self._sample_n_times



import copy
import datetime
from collections.abc import Iterable, Mapping, Sequence
from math import ceil
from pathlib import Path

import cftime
import numpy as np
import torch
import xarray as xr
from netCDF4 import Dataset

from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.utils import (
    DIM_INFO_HEALPIX,
    DIM_INFO_LATLON,
    get_all_names,
)
from fme.core.dataset.data_typing import VariableMetadata

LEAD_TIME_DIM = "time"
LEAD_TIME_UNITS = "months"
ENSEMBLE_DIM = "sample"
INIT_TIME = "init_time"
VALID_TIME = "valid_time"
TIME_UNITS = "days since 1970-01-01 00:00:00"
COUNTS = "counts"


def months_for_timesteps(n_timesteps: int, timestep: datetime.timedelta) -> int:
    steps_per_day = datetime.timedelta(days=1) / timestep
    return ceil(n_timesteps * (12.0 / (365.24 * steps_per_day))) + 2


class PairedMonthlyDataWriter:
    """
    Wrapper over MonthlyDataWriter to write both target and prediction data
    to the same file.

    Gives the same interface as for our other writers.
    """

    def __init__(
        self,
        path: str,
        n_samples: int,
        n_timesteps: int,
        timestep: datetime.timedelta,
        save_names: Sequence[str] | None,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ):
        n_months = months_for_timesteps(n_timesteps, timestep)
        self._target_writer = MonthlyDataWriter(
            path=path,
            label="target",
            n_samples=n_samples,
            n_months=n_months,
            save_names=save_names,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=dataset_metadata,
        )
        self._prediction_writer = MonthlyDataWriter(
            path=path,
            label="predictions",
            n_samples=n_samples,
            n_months=n_months,
            save_names=save_names,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=dataset_metadata,
        )

    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        self._target_writer.append_batch(
            data=target, start_timestep=start_timestep, batch_time=batch_time
        )
        self._prediction_writer.append_batch(
            data=prediction, start_timestep=start_timestep, batch_time=batch_time
        )

    def flush(self):
        self._target_writer.flush()
        self._prediction_writer.flush()


class MonthlyDataWriter:
    """
    Write monthly total data and sample counts to a netCDF file.

    Allows computing the mean afterwards by dividing the total by the counts.
    """

    def __init__(
        self,
        path: str,
        label: str,
        n_samples: int,
        n_months: int,
        save_names: Sequence[str] | None,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ):
        """
        Args:
            path: Directory to write netCDF file(s).
            label: Label to append to the filename.
            n_samples: Number of samples to write to the file, each sample being
                an ensemble member.
            n_months: Number of months to write to the file.
            save_names: Names of variables to save in the predictions netcdf file.
                If None, all predicted variables will be saved.
            variable_metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
            dataset_metadata: Metadata for the dataset.
        """
        filename = str(Path(path) / f"monthly_mean_{label}.nc")
        self._save_names = save_names
        self.variable_metadata = variable_metadata
        self.coords = coords
        self.dataset = Dataset(filename, "w", format="NETCDF4")
        self.dataset.createDimension(LEAD_TIME_DIM, n_months)
        self.dataset.createVariable(LEAD_TIME_DIM, "i8", (LEAD_TIME_DIM,))
        self.dataset.variables[LEAD_TIME_DIM].units = LEAD_TIME_UNITS
        self.dataset.variables[LEAD_TIME_DIM][:] = np.arange(n_months)
        self.dataset.createDimension(ENSEMBLE_DIM, n_samples)
        self.dataset.createVariable(INIT_TIME, "i8", (ENSEMBLE_DIM,))
        self.dataset.variables[INIT_TIME].units = TIME_UNITS
        self.dataset.createVariable(COUNTS, "i8", (ENSEMBLE_DIM, LEAD_TIME_DIM))
        self.dataset.createVariable(
            VALID_TIME,
            "i8",
            (
                ENSEMBLE_DIM,
                LEAD_TIME_DIM,
            ),
        )
        self.dataset.variables[VALID_TIME].units = TIME_UNITS
        self.dataset.variables[COUNTS][:] = 0
        dataset_metadata = copy.copy(dataset_metadata)
        dataset_metadata.title = f"ACE monthly {label} data file"
        for key, value in dataset_metadata.as_flat_str_dict().items():
            self.dataset.setncattr(key, value)
        self._init_years = np.full([n_samples], -1, dtype=int)
        self._init_months = np.full([n_samples], -1, dtype=int)
        self._dataset_dims_created = False

    def _get_initial_year_and_month(
        self,
        years: np.ndarray,
        months: np.ndarray,
        calendar: str,
    ) -> tuple[np.ndarray, np.ndarray]:
        reference_date = cftime.datetime(1970, 1, 1, calendar=calendar)
        if self._init_years[0] == -1:
            self._init_years[:] = years
            self._init_months[:] = months
            n_months = self.dataset.variables[LEAD_TIME_DIM].shape[0]
            days_since_reference = get_days_since_reference(
                years=years,
                months=months,
                n_months=n_months,
                reference_date=reference_date,
                calendar=calendar,
            )
            self.dataset.variables[INIT_TIME][:] = days_since_reference[:, 0]
            # use the 15th of each month, which is 14 days into the month
            self.dataset.variables[VALID_TIME][:, :] = days_since_reference + 14
        return (self._init_years, self._init_months)

    def _get_month_indices(self, batch_time: xr.DataArray) -> np.ndarray:
        """
        Get the month indices for the batch of data.

        If this is the first time called, the times are assumed to be the
        initial times for the data, and are stored for determining the month
        indices in this and future calls.

        Args:
            batch_time: Time coordinate for each sample in the batch, of shape
                [ensemble_member, lead_time].

        Returns:
            Month indices for the batch of data.
        """
        calendar = batch_time.dt.calendar
        years = batch_time.dt.year.values
        # datetime months are 1-indexed, we want 0-indexed
        months = batch_time.dt.month.values - 1
        init_years, init_months = self._get_initial_year_and_month(
            years=years[:, 0], months=months[:, 0], calendar=calendar
        )
        return 12 * (years - init_years[:, None]) + (months - init_months[:, None])

    def _get_variable_names_to_save(
        self, *data_varnames: Iterable[str]
    ) -> Iterable[str]:
        return get_all_names(*data_varnames, allowlist=self._save_names)

    def append_batch(
        self,
        data: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            data: Values to store.
            start_timestep: Timestep index for the start of the batch, unused.
            batch_time: Time coordinate for each sample in the batch.
        """
        del start_timestep  # unused
        n_samples_data = list(data.values())[0].shape[0]
        n_samples_time = batch_time.sizes["sample"]
        if n_samples_data != n_samples_time:
            raise ValueError(
                f"Batch size mismatch, data has {n_samples_data} samples "
                f"and batch_time has {n_samples_time} samples."
            )
        n_times_data = list(data.values())[0].shape[1]
        n_times_time = batch_time.sizes["time"]
        if n_times_data != n_times_time:
            raise ValueError(
                f"Batch time dimension mismatch, data has {n_times_data} times "
                f"and batch_time has {n_times_time} times."
            )

        if not self._dataset_dims_created:
            _dim_info = DIM_INFO_HEALPIX if "face" in self.coords else DIM_INFO_LATLON
            _ordered_names = []
            for dim in _dim_info:
                dim_size = data[next(iter(data.keys()))].shape[dim.index]
                self.dataset.createDimension(dim.name, dim_size)
                if dim.name in self.coords:
                    self.dataset.createVariable(dim.name, "f4", (dim.name,))
                    self.dataset.variables[dim.name][:] = self.coords[dim.name]
                _ordered_names.append(dim.name)
            dims = (ENSEMBLE_DIM, LEAD_TIME_DIM, *_ordered_names)
            self._dataset_dims_created = True

        save_names = self._get_variable_names_to_save(data.keys())
        months = self._get_month_indices(batch_time)
        month_min = np.min(months)
        month_range = np.max(months) - month_min + 1
        count_data = self.dataset.variables[COUNTS][
            :, month_min : month_min + month_range
        ]
        for variable_name in save_names:
            # define the variable if it doesn't exist
            if variable_name not in self.dataset.variables:
                self.dataset.createVariable(
                    variable_name,
                    "f4",
                    dims,
                    fill_value=np.nan,
                )
                self.dataset.variables[variable_name][:] = 0.0
                if variable_name in self.variable_metadata:
                    self.dataset.variables[
                        variable_name
                    ].units = self.variable_metadata[variable_name].units
                    self.dataset.variables[
                        variable_name
                    ].long_name = self.variable_metadata[variable_name].long_name
                self.dataset.variables[variable_name].coordinates = " ".join(
                    [INIT_TIME, VALID_TIME, COUNTS]
                )

            array = data[variable_name].cpu().numpy()

            # Add the data to the variable totals
            # Have to extract the data and write it back as `.at` does not play nicely
            # with netCDF4
            # We pull just the month subset we need for speed reasons
            month_data = self.dataset.variables[variable_name][
                :, month_min : month_min + month_range
            ]
            add_data(
                target=month_data,
                target_start_counts=count_data,
                source=array,
                months_elapsed=months - month_min,
            )
            self.dataset.variables[variable_name][
                :, month_min : month_min + month_range
            ] = month_data
        # counts must be added after data, as we use the base counts when updating means
        for i_sample in range(n_samples_data):
            self.dataset.variables[COUNTS][i_sample] += np.bincount(
                months[i_sample], minlength=self.dataset.variables[COUNTS].shape[1]
            )

        self.dataset.sync()

    def flush(self):
        """
        Flush the data to disk.
        """
        self.dataset.sync()


def add_data(
    *,
    target: np.ndarray,
    target_start_counts: np.ndarray,
    source: np.ndarray,
    months_elapsed: np.ndarray,
):
    """
    Add source data to target monthly mean data, aggregating by month.

    All operations are performed independently on each batch member [b, ...].

    Args:
        target: Array of monthly mean data to add to, of shape
            [b, month].
        target_start_counts: Array of counts for each month, of shape
            [b, month]. This array does not get updated.
        source: Array of values to add into the monthly aggregates, of shape [b, time].
        months_elapsed: Elapsed months of source since the start of the data,
            of shape [b, time],
            corresponding to an index of the target array for each value in source.
            Assumed to be monotonically increasing.
    """
    for i_sample in range(source.shape[0]):
        i_time = 0
        while i_time < source.shape[1]:
            month_index = months_elapsed[i_sample, i_time]
            i_month_boundary = i_time + find_boundary(
                months_elapsed[i_sample, i_time:], month_index
            )
            # Calculate sum of new data for the current month
            new_data_sum = np.sum(source[i_sample, i_time:i_month_boundary], axis=0)
            new_samples_count = i_month_boundary - i_time

            # Update target mean for the month
            old_mean = target[i_sample, month_index]
            old_count = target_start_counts[i_sample, month_index]
            new_mean = (old_mean * old_count + new_data_sum) / (
                old_count + new_samples_count
            )

            target[i_sample, month_index] = new_mean

            i_time = i_month_boundary


def find_boundary(month_array, start_month) -> int:
    """
    Assuming month_array is an ordered array of months,
    find the index of the first month that is not start_month.
    """
    return np.searchsorted(month_array, start_month, side="right")


def get_days_since_reference(
    years: np.ndarray,
    months: np.ndarray,
    reference_date: cftime.datetime,
    n_months: int,
    calendar: str,
) -> np.ndarray:
    """
    Get the days since a reference date for each month.

    Args:
        years: Array of years, of shape [n_samples].
        months: Array of months, of shape [n_samples], zero-indexed.
        reference_date: Reference date for the calendar.
        n_months: Number of months to compute starting at each sample (year, month).
        calendar: Calendar to use.
    """
    months_elapsed = np.arange(n_months)
    calendar_month = (months[:, None] + months_elapsed[None, :]) % 12
    calendar_year = years[:, None] + (months[:, None] + months_elapsed[None, :]) // 12
    days_since_reference = np.zeros_like(calendar_month, dtype=np.int64)
    for i in range(calendar_month.shape[0]):
        dates_sample = xr.date_range(
            cftime.datetime(
                calendar_year[i, 0], calendar_month[i, 0] + 1, 1, calendar=calendar
            ),
            cftime.datetime(
                calendar_year[i, -1], calendar_month[i, -1] + 1, 1, calendar=calendar
            ),
            freq="MS",
            calendar=calendar,
            use_cftime=True,
        )
        days_since_reference[i, :] = (
            dates_sample.values - reference_date
        ) // datetime.timedelta(days=1)
    return days_since_reference



import cftime
import xarray as xr


def date_range(
    n_timesteps: int,
    start_time: cftime.datetime = cftime.DatetimeGregorian(2020, 1, 1, 0, 0, 0),
    freq: str = "6h",
) -> xr.DataArray:
    """
    Create a DataArray with a time dimension of shape (n_timesteps,).
    """
    return xr.DataArray(
        xr.date_range(
            start_time,
            freq=freq,
            periods=n_timesteps,
            use_cftime=True,
        ).values,
        dims="time",
    )



import dataclasses
import pathlib
import tempfile

import dacite
import pytest
import torch

from fme.ace.testing.fv3gfs_data import get_scalar_dataset
from fme.core.device import move_tensordict_to_device
from fme.core.normalizer import (
    NetworkAndLossNormalizationConfig,
    NormalizationConfig,
    StandardNormalizer,
    _combine_normalizers,
)


def test_normalize_depends_on_mean():
    means = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalized = normalizer.normalize(tensors)
    assert normalized["a"] == torch.tensor(0.0)
    assert normalized["b"] == torch.tensor(-1.0)


def test_normalize_depends_on_std():
    means = {"a": torch.tensor(0.0), "b": torch.tensor(0.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalized = normalizer.normalize(tensors)
    assert normalized["a"] == torch.tensor(1.0)
    assert normalized["b"] == torch.tensor(0.5)


def test_denormalize_depends_on_mean():
    means = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(1.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(0.0), "b": torch.tensor(-1.0)}
    denormalized = normalizer.denormalize(tensors)
    assert denormalized["a"] == torch.tensor(1.0)
    assert denormalized["b"] == torch.tensor(1.0)


def test_denormalize_depends_on_std():
    means = {"a": torch.tensor(0.0), "b": torch.tensor(0.0)}
    stds = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = {"a": torch.tensor(1.0), "b": torch.tensor(0.5)}
    denormalized = normalizer.denormalize(tensors)
    assert denormalized["a"] == torch.tensor(1.0)
    assert denormalized["b"] == torch.tensor(1.0)


def test_normalize_and_denormalize_random_tensor():
    torch.manual_seed(0)
    # randomly set means and stds
    means = move_tensordict_to_device({"a": torch.randn(1), "b": torch.randn(1)})
    stds = move_tensordict_to_device({"a": torch.randn(1), "b": torch.randn(1)})
    normalizer = StandardNormalizer(means=means, stds=stds)
    tensors = move_tensordict_to_device({"a": torch.randn(10), "b": torch.randn(10)})
    denormalized = normalizer.denormalize(normalizer.normalize(tensors))
    torch.testing.assert_close(denormalized["a"], tensors["a"])
    torch.testing.assert_close(denormalized["b"], tensors["b"])


def test_missing_normalization_build_raises_error():
    normalization = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 1.0},
    )
    all_names = ["a", "b", "c"]
    with pytest.raises(KeyError):
        normalization.build(all_names)


def test_tensors_with_missing_normalization_stats_get_filtered():
    normalization = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 1.0},
    ).build(["a", "b"])
    sample_input = {"a": torch.zeros(1), "b": torch.zeros(1), "c": torch.zeros(1)}
    sample_input = move_tensordict_to_device(sample_input)

    normalized = normalization.normalize(sample_input)
    assert "c" not in normalized

    denormalized = normalization.denormalize(sample_input)
    assert "c" not in denormalized


@pytest.mark.parametrize("fill_nans_on_normalize", [True, False])
@pytest.mark.parametrize("fill_nans_on_denormalize", [True, False])
def test_normalization_with_nans(fill_nans_on_normalize, fill_nans_on_denormalize):
    means = {"a": 1.0, "b": 2.0}
    stds = {"a": 1.0, "b": 2.0}
    normalization = NormalizationConfig(
        means=means,
        stds=stds,
        fill_nans_on_normalize=fill_nans_on_normalize,
        fill_nans_on_denormalize=fill_nans_on_denormalize,
    ).build(["a", "b"])
    denormalized_input = {
        "a": torch.tensor([-1.0, float("nan"), 1.0]),
        "b": torch.tensor([0.0, float("nan"), 4.0]),
    }
    denormalized_input = move_tensordict_to_device(denormalized_input)
    normalized = normalization.normalize(denormalized_input)
    if fill_nans_on_normalize:
        assert not torch.isnan(normalized["a"][1]), "normalized_nans_removed_a"
        assert normalized["a"][1] == torch.tensor(0), "normalized_nans_filled_means_a"
        assert not torch.isnan(normalized["b"][1]), "normalized_nans_removed_b"
        assert normalized["b"][1] == torch.tensor(0), "normalized_nans_filled_means_b"
    else:
        assert torch.isnan(normalized["a"][1]), "normalized_nans_not_removed_a"
        assert torch.isnan(normalized["b"][1]), "normalized_nans_not_removed_b"

    normalized_input = {
        "a": torch.tensor([-1.0, float("nan"), 1.0]),
        "b": torch.tensor([-1.0, float("nan"), 1.0]),
    }
    normalized_input = move_tensordict_to_device(normalized_input)
    denormalized = normalization.denormalize(normalized_input)
    if fill_nans_on_denormalize:
        assert not torch.isnan(denormalized["a"][1]), "denormalized_nans_removed_a"
        assert denormalized["a"][1] == torch.tensor(
            means["a"]
        ), "denormalized_nans_filled_means_a"
        assert not torch.isnan(denormalized["b"][1]), "denormalized_nans_removed_b"
        assert denormalized["b"][1] == torch.tensor(
            means["b"]
        ), "denormalized_nans_filled_means_b"
    else:
        assert torch.isnan(denormalized["a"][1]), "denormalized_nans_not_removed_a"
        assert torch.isnan(denormalized["b"][1]), "denormalized_nans_not_removed_b"


def test_combined_normalization_uses_network_normalizer_for_loss():
    torch.manual_seed(0)
    network_config = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 2.0},
    )
    combined_config = NetworkAndLossNormalizationConfig(
        network=network_config,
    )
    direct_normalizer = network_config.build(["a", "b"])
    loss_normalizer = combined_config.get_loss_normalizer(
        names=["a", "b"],
        residual_scaled_names=["a", "b"],
    )
    data = move_tensordict_to_device({"a": torch.randn(10), "b": torch.randn(10)})
    direct_normalized = direct_normalizer.normalize(data)
    loss_normalized = loss_normalizer.normalize(data)
    torch.testing.assert_close(direct_normalized["a"], loss_normalized["a"])
    torch.testing.assert_close(direct_normalized["b"], loss_normalized["b"])


@pytest.mark.parametrize("are_prognostic", [True, False])
def test_combined_normalization_uses_loss_normalizer_for_loss(are_prognostic: bool):
    torch.manual_seed(0)
    network_config = NormalizationConfig(
        means={"a": torch.randn(1), "b": torch.randn(1)},
        stds={"a": torch.randn(1), "b": torch.randn(1)},
    )
    loss_config = NormalizationConfig(
        means={"a": torch.randn(1), "b": torch.randn(1)},
        stds={"a": torch.randn(1), "b": torch.randn(1)},
    )
    combined_config = NetworkAndLossNormalizationConfig(
        network=network_config,
        loss=loss_config,
    )
    direct_normalizer = loss_config.build(["a", "b"])
    if are_prognostic:
        prognostic_names = ["a", "b"]
    else:
        prognostic_names = []
    loss_normalizer = combined_config.get_loss_normalizer(
        names=["a", "b"],
        residual_scaled_names=prognostic_names,
    )
    data = move_tensordict_to_device({"a": torch.randn(10), "b": torch.randn(10)})
    direct_normalized = direct_normalizer.normalize(data)
    loss_normalized = loss_normalizer.normalize(data)
    torch.testing.assert_close(direct_normalized["a"], loss_normalized["a"])
    torch.testing.assert_close(direct_normalized["b"], loss_normalized["b"])


def test_combined_normalization_uses_residual_normalizer_for_prognostic_loss():
    torch.manual_seed(0)
    network_config = NormalizationConfig(
        means={"a": torch.randn(1), "b": torch.randn(1)},
        stds={"a": torch.randn(1), "b": torch.randn(1)},
    )
    residual_config = NormalizationConfig(
        means={"a": torch.randn(1), "b": torch.randn(1)},
        stds={"a": torch.randn(1), "b": torch.randn(1)},
    )
    combined_config = NetworkAndLossNormalizationConfig(
        network=network_config,
        residual=residual_config,
    )
    direct_residual_normalizer = residual_config.build(["a", "b"])
    direct_network_normalizer = network_config.build(["a", "b"])
    loss_normalizer = combined_config.get_loss_normalizer(
        names=["a", "b"],
        residual_scaled_names=["a"],
    )
    data = move_tensordict_to_device({"a": torch.randn(10), "b": torch.randn(10)})
    direct_residual_normalized = direct_residual_normalizer.normalize(data)
    direct_network_noramlized = direct_network_normalizer.normalize(data)
    loss_normalized = loss_normalizer.normalize(data)
    torch.testing.assert_close(direct_residual_normalized["a"], loss_normalized["a"])
    torch.testing.assert_close(direct_network_noramlized["b"], loss_normalized["b"])


def test_combined_normalization_cannot_set_both_loss_and_residual():
    network_config = NormalizationConfig(
        means={"a": torch.randn(1), "b": torch.randn(1)},
        stds={"a": torch.randn(1), "b": torch.randn(1)},
    )
    with pytest.raises(ValueError):
        NetworkAndLossNormalizationConfig(
            network=network_config,
            loss=network_config,
            residual=network_config,
        )


def test_normalization_config_with_means_and_stds_round_trip():
    config = NormalizationConfig(
        means={"a": 1.0, "b": 2.0},
        stds={"a": 1.0, "b": 2.0},
    )
    round_tripped = dacite.from_dict(
        NormalizationConfig,
        data=dataclasses.asdict(config),
        config=dacite.Config(
            strict=True,
        ),
    )
    assert config == round_tripped


def test__combine_normalizers():
    vars = ["prog_0", "prog_1", "diag_0"]
    full_field_normalizer = StandardNormalizer(
        means={var: torch.rand(3) for var in vars},
        stds={var: torch.rand(3) for var in vars},
        fill_nans_on_normalize=True,
        fill_nans_on_denormalize=True,
    )
    residual_normalizer = StandardNormalizer(
        means={var: torch.rand(3) for var in ["prog_0", "prog_1"]},
        stds={var: torch.rand(3) for var in ["prog_0", "prog_1"]},
    )
    combined_normalizer = _combine_normalizers(
        override_normalizer=residual_normalizer,
        base_normalizer=full_field_normalizer,
    )
    assert combined_normalizer.fill_nans_on_normalize
    assert combined_normalizer.fill_nans_on_denormalize
    for var in combined_normalizer.means:
        if "prog" in var:
            assert torch.allclose(
                combined_normalizer.means[var], residual_normalizer.means[var]
            )
            assert torch.allclose(
                combined_normalizer.stds[var], residual_normalizer.stds[var]
            )
        else:
            assert torch.allclose(
                combined_normalizer.means[var], full_field_normalizer.means[var]
            )
            assert torch.allclose(
                combined_normalizer.stds[var], full_field_normalizer.stds[var]
            )


def test_build_from_files():
    mean_ds = get_scalar_dataset(["a", "b", "c"], fill_value=1.0)
    std_ds = get_scalar_dataset(["a", "b", "c"], fill_value=2.0)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        mean_ds.to_netcdf(tmp_path / "mean.nc")
        std_ds.to_netcdf(tmp_path / "std.nc")
        normalizer = NormalizationConfig(
            global_means_path=tmp_path / "mean.nc",
            global_stds_path=tmp_path / "std.nc",
        ).build(["a", "b"])
        for name in ["a", "b"]:
            assert normalizer.means[name] == 1.0
            assert normalizer.stds[name] == 2.0
        assert "c" not in normalizer.means
        assert "c" not in normalizer.stds


@pytest.mark.parametrize("fill_nans_on_normalize", [True, False])
@pytest.mark.parametrize("fill_nans_on_denormalize", [True, False])
def test_load_from_files(fill_nans_on_normalize: bool, fill_nans_on_denormalize: bool):
    mean_ds = get_scalar_dataset(["a", "b", "c"], fill_value=1.0)
    std_ds = get_scalar_dataset(["a", "b", "c"], fill_value=2.0)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        mean_ds.to_netcdf(tmp_path / "mean.nc")
        std_ds.to_netcdf(tmp_path / "std.nc")
        config = NormalizationConfig(
            global_means_path=tmp_path / "mean.nc",
            global_stds_path=tmp_path / "std.nc",
            fill_nans_on_normalize=fill_nans_on_normalize,
            fill_nans_on_denormalize=fill_nans_on_denormalize,
        )
        config.load()
    assert config.fill_nans_on_normalize == fill_nans_on_normalize
    assert config.fill_nans_on_denormalize == fill_nans_on_denormalize
    normalizer = config.build(["a", "b"])
    for name in ["a", "b"]:
        assert normalizer.means[name] == 1.0
        assert normalizer.stds[name] == 2.0
    assert "c" not in normalizer.means
    assert "c" not in normalizer.stds


def test_cannot_build_without_load_or_files():
    mean_ds = get_scalar_dataset(["a", "b", "c"], fill_value=1.0)
    std_ds = get_scalar_dataset(["a", "b", "c"], fill_value=2.0)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        mean_ds.to_netcdf(tmp_path / "mean.nc")
        std_ds.to_netcdf(tmp_path / "std.nc")
        config = NormalizationConfig(
            global_means_path=tmp_path / "mean.nc",
            global_stds_path=tmp_path / "std.nc",
        )
    with pytest.raises(FileNotFoundError):
        config.build(["a", "b"])


def test_cannot_load_without_files():
    mean_ds = get_scalar_dataset(["a", "b", "c"], fill_value=1.0)
    std_ds = get_scalar_dataset(["a", "b", "c"], fill_value=2.0)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)
        mean_ds.to_netcdf(tmp_path / "mean.nc")
        std_ds.to_netcdf(tmp_path / "std.nc")
        config = NormalizationConfig(
            global_means_path=tmp_path / "mean.nc",
            global_stds_path=tmp_path / "std.nc",
        )
    with pytest.raises(FileNotFoundError):
        config.load()


def test_can_create_config_without_files():
    NormalizationConfig(
        global_means_path="/not/a/real/path",
        global_stds_path="/not/a/real/path",
    )



import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.aggregator.one_step import OneStepAggregator
from fme.ace.stepper import TrainOutput
from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.typing_ import EnsembleTensorDict


def get_ds_info(nx: int, ny: int) -> DatasetInfo:
    coords = LatLonCoordinates(lon=torch.arange(nx), lat=torch.arange(ny))
    return DatasetInfo(horizontal_coordinates=coords)


def test_labels_exist():
    batch_size = 10
    n_ensemble = 2
    n_time = 3
    nx, ny = 2, 2
    loss = 1.0
    ds_info = get_ds_info(nx, ny)
    agg = OneStepAggregator(ds_info, save_diagnostics=False)
    target_data = EnsembleTensorDict(
        {"a": torch.randn(batch_size, 1, n_time, nx, ny, device=get_device())},
    )
    gen_data = EnsembleTensorDict(
        {"a": torch.randn(batch_size, n_ensemble, n_time, nx, ny, device=get_device())},
    )
    agg.record_batch(
        batch=TrainOutput(
            metrics={"loss": loss},
            target_data=target_data,
            gen_data=gen_data,
            time=xr.DataArray(np.zeros((batch_size, n_time)), dims=["sample", "time"]),
            normalize=lambda x: x,
        ),
    )
    logs = agg.get_logs(label="test")
    expected_keys = [
        "test/mean/loss",
        "test/mean/weighted_rmse/a",
        "test/mean/weighted_bias/a",
        "test/mean/weighted_grad_mag_percent_diff/a",
        "test/snapshot/image-full-field/a",
        "test/snapshot/image-residual/a",
        "test/snapshot/image-error/a",
        "test/mean_map/image-full-field/a",
        "test/mean_map/image-error/a",
        "test/power_spectrum/positive_norm_bias/a",
        "test/power_spectrum/negative_norm_bias/a",
        "test/power_spectrum/mean_abs_norm_bias/a",
        "test/power_spectrum/smallest_scale_norm_bias/a",
        "test/crps/a",
        "test/crps/mean_map/a",
        "test/ssr_bias/a",
        "test/ssr_bias/mean_map/a",
    ]
    assert set(logs.keys()) == set(expected_keys)


def test_aggregator_raises_on_no_data():
    """
    Basic test the aggregator combines loss correctly
    with multiple batches and no distributed training.
    """
    ds_info = get_ds_info(2, 2)
    agg = OneStepAggregator(ds_info, save_diagnostics=False)
    with pytest.raises(ValueError) as excinfo:
        agg.record_batch(
            batch=TrainOutput(
                metrics={"loss": 1.0},
                target_data=EnsembleTensorDict({}),
                gen_data=EnsembleTensorDict({}),
                time=xr.DataArray(np.zeros((0, 0)), dims=["sample", "time"]),
                normalize=lambda x: x,
            ),
        )
        # check that the raised exception contains the right substring
        assert "No data" in str(excinfo.value)


@pytest.mark.parametrize(
    "epoch", [pytest.param(None, id="no epoch"), pytest.param(2, id="epoch 2")]
)
def test_flush_diagnostics(tmpdir, epoch):
    nx, ny, batch_size, n_ensemble, n_time = 3, 3, 10, 2, 3
    ds_info = get_ds_info(nx, ny)
    agg = OneStepAggregator(ds_info, output_dir=(tmpdir / "val"))
    target_data = EnsembleTensorDict(
        {"a": torch.randn(batch_size, 1, n_time, nx, ny, device=get_device())}
    )
    gen_data = EnsembleTensorDict(
        {"a": torch.randn(batch_size, n_ensemble, n_time, nx, ny, device=get_device())}
    )
    time = xr.DataArray(np.zeros((batch_size, n_time)), dims=["sample", "time"])
    agg.record_batch(
        batch=TrainOutput(
            metrics={"loss": 1.0},
            target_data=target_data,
            gen_data=gen_data,
            time=time,
            normalize=lambda x: x,
        ),
    )
    if epoch is not None:
        agg.flush_diagnostics(subdir=f"epoch_{epoch:04d}")
        output_dir = tmpdir / "val" / f"epoch_{epoch:04d}"
    else:
        agg.flush_diagnostics()
        output_dir = tmpdir / "val"
    expected_files = [
        "mean",
        "snapshot",
        "mean_map",
    ]
    for file in expected_files:
        assert (output_dir / f"{file}_diagnostics.nc").exists()


def test_agg_raises_without_output_dir():
    ds_info = get_ds_info(nx=2, ny=2)
    with pytest.raises(
        ValueError, match="Output directory must be set to save diagnostics"
    ):
        OneStepAggregator(ds_info, save_diagnostics=True, output_dir=None)



import dataclasses
from collections.abc import Mapping
from typing import Any

import numpy as np
import torch
import torch.nn as nn

from fme.ace.models.ocean.m2lines.layers import AvgPool, BilinearUpsample, ConvNeXtBlock
from fme.ace.models.ocean.m2lines.utils import pairwise


class Samudra(torch.nn.Module):
    """
    Samudra Network from M2Lines.

    Parameters
    ----------
    input_channels : int
        Number of input channels, including forcing variables and history
    output_channels : int
        Number of output channels in the final layer
    ch_width : List[int]
        Channel widths for each level of the U-Net architecture
    dilation : List[int]
        Dilation rates for each ConvNeXt block
    n_layers : List[int]
        Number of ConvNeXt layers at each level
    pad : str, optional
        Type of padding to use in convolutions, for example,
        ('circular', 'constant'), by default "circular"
    norm: str, optional
        Normalization to use in the network, by default "instance"
        Options are "batch", "layer", "instance", or None
        "layer" normalization normalizes over only the channel dimensions

    Example:
    --------
    >>> import torch
    >>> from fme.ace.models.ocean.m2lines.samudra import Samudra
    >>> model = Samudra(
    ...     input_channels=4,
    ...     output_channels=3,
    ...     ch_width=[8],
    ...     dilation=[2],
    ...     n_layers=[1],
    ... )
    >>> model(torch.randn(1, 4, 128, 128)).shape
    torch.Size([1, 3, 128, 128])
    """

    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        ch_width: list[int] = dataclasses.field(
            default_factory=lambda: [200, 250, 300, 400]
        ),
        dilation: list[int] = dataclasses.field(default_factory=lambda: [1, 2, 4, 8]),
        n_layers: list[int] = dataclasses.field(default_factory=lambda: [1, 1, 1, 1]),
        pad: str = "circular",
        norm: str | None = "instance",
        norm_kwargs: Mapping[str, Any] | None = None,
    ):
        super().__init__()

        self.input_channels = input_channels
        self.output_channels = output_channels
        self.hist = 0  # Fixed
        self.ch_width = ch_width
        self.dilation = dilation
        self.n_layers = n_layers
        self.pad = pad
        self.norm = norm
        self.norm_kwargs = norm_kwargs
        self.last_kernel_size = 3
        self.N_pad = int((self.last_kernel_size - 1) / 2)

        ch_width_with_input = (self.input_channels, *self.ch_width)

        # going down
        layers = []
        for i, (a, b) in enumerate(pairwise(ch_width_with_input)):
            layers.append(
                ConvNeXtBlock(
                    a,
                    b,
                    dilation=self.dilation[i],
                    n_layers=self.n_layers[i],
                    pad=self.pad,
                    norm=self.norm,
                    norm_kwargs=self.norm_kwargs,
                )
            )
            layers.append(AvgPool())
        layers.append(
            ConvNeXtBlock(
                b,
                b,
                dilation=self.dilation[i],
                n_layers=self.n_layers[i],
                pad=self.pad,
                norm=self.norm,
                norm_kwargs=self.norm_kwargs,
            )
        )
        layers.append(BilinearUpsample(in_channels=b, out_channels=b))
        ch_width_with_input_reversed = ch_width_with_input[::-1]
        dilation_reversed = self.dilation[::-1]
        n_layers_reversed = self.n_layers[::-1]
        for i, (a, b) in enumerate(pairwise(ch_width_with_input_reversed[:-1])):
            layers.append(
                ConvNeXtBlock(
                    a,
                    b,
                    dilation=dilation_reversed[i],
                    n_layers=n_layers_reversed[i],
                    pad=self.pad,
                    norm=self.norm,
                    norm_kwargs=self.norm_kwargs,
                )
            )
            layers.append(BilinearUpsample(in_channels=b, out_channels=b))
        layers.append(
            ConvNeXtBlock(
                b,
                b,
                dilation=dilation_reversed[i],
                n_layers=n_layers_reversed[i],
                pad=self.pad,
                norm=self.norm,
                norm_kwargs=self.norm_kwargs,
            )
        )
        layers.append(torch.nn.Conv2d(b, self.output_channels, self.last_kernel_size))

        self.layers = nn.ModuleList(layers)
        self.num_steps = int(len(ch_width_with_input) - 1)

    def forward(self, fts):
        temp: list[torch.Tensor] = []
        count = 0
        for layer in self.layers:
            crop = fts.shape[2:]
            if isinstance(layer, nn.Conv2d):
                fts = torch.nn.functional.pad(
                    fts, (self.N_pad, self.N_pad, 0, 0), mode=self.pad
                )
                fts = torch.nn.functional.pad(
                    fts, (0, 0, self.N_pad, self.N_pad), mode="constant"
                )
            fts = layer(fts)
            if count < self.num_steps:
                if isinstance(layer, ConvNeXtBlock):
                    temp.append(fts)
                    count += 1
            elif count >= self.num_steps:
                if isinstance(layer, BilinearUpsample):
                    crop = np.array(fts.shape[2:])
                    shape = np.array(
                        temp[int(2 * self.num_steps - count - 1)].shape[2:]
                    )
                    pads = shape - crop
                    pads = [
                        pads[1] // 2,
                        pads[1] - pads[1] // 2,
                        pads[0] // 2,
                        pads[0] - pads[0] // 2,
                    ]
                    fts = nn.functional.pad(fts, pads)
                    fts += temp[int(2 * self.num_steps - count - 1)]
                    count += 1
        return fts



import dataclasses
import datetime
import warnings
from collections.abc import Mapping, Sequence
from pathlib import Path

import numpy as np
import torch
import xarray as xr

from fme.ace.data_loading.batch_data import BatchData, PairedData, PrognosticState
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.generics.writer import WriterABC

from .dataset_metadata import DatasetMetadata
from .histograms import PairedHistogramDataWriter
from .monthly import MonthlyDataWriter, PairedMonthlyDataWriter, months_for_timesteps
from .raw import PairedRawDataWriter, RawDataWriter
from .time_coarsen import PairedTimeCoarsen, TimeCoarsen, TimeCoarsenConfig
from .video import PairedVideoDataWriter

PairedSubwriter = (
    PairedRawDataWriter
    | PairedVideoDataWriter
    | PairedHistogramDataWriter
    | PairedTimeCoarsen
    | PairedMonthlyDataWriter
)

Subwriter = MonthlyDataWriter | RawDataWriter | TimeCoarsen


@dataclasses.dataclass
class DataWriterConfig:
    """
    Configuration for inference data writers.

    Parameters:
        log_extended_video_netcdfs: Whether to enable writing of netCDF files
            containing video metrics.
        save_prediction_files: Whether to enable writing of netCDF files
            containing the predictions and target values.
        save_monthly_files: Whether to enable writing of netCDF files
            containing the monthly predictions and target values.
        names: Names of variables to save in the prediction, histogram, and monthly
            netCDF files.
        save_histogram_files: Enable writing of netCDF files containing histograms.
        time_coarsen: Configuration for time coarsening of written outputs.
    """

    log_extended_video_netcdfs: bool = False
    save_prediction_files: bool = True
    save_monthly_files: bool = True
    names: Sequence[str] | None = None
    save_histogram_files: bool = False
    time_coarsen: TimeCoarsenConfig | None = None

    def __post_init__(self):
        if (
            not any(
                [
                    self.save_prediction_files,
                    self.save_monthly_files,
                    self.save_histogram_files,
                ]
            )
            and self.names is not None
        ):
            warnings.warn(
                "names provided but all options to "
                "save subsettable output files are False."
            )

    def build_paired(
        self,
        experiment_dir: str,
        n_initial_conditions: int,
        n_timesteps: int,
        timestep: datetime.timedelta,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ) -> "PairedDataWriter":
        return PairedDataWriter(
            path=experiment_dir,
            n_initial_conditions=n_initial_conditions,
            n_timesteps=n_timesteps,
            timestep=timestep,
            variable_metadata=variable_metadata,
            coords=coords,
            enable_prediction_netcdfs=self.save_prediction_files,
            enable_monthly_netcdfs=self.save_monthly_files,
            enable_video_netcdfs=self.log_extended_video_netcdfs,
            save_names=self.names,
            enable_histogram_netcdfs=self.save_histogram_files,
            time_coarsen=self.time_coarsen,
            dataset_metadata=dataset_metadata,
        )

    def build(
        self,
        experiment_dir: str,
        n_initial_conditions: int,
        n_timesteps: int,
        timestep: datetime.timedelta,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ) -> "DataWriter":
        if self.save_histogram_files:
            raise NotImplementedError(
                "Saving histograms is not supported for prediction-only data writers. "
                "Make sure to set `save_histogram_files=False`."
            )
        if self.log_extended_video_netcdfs:
            raise NotImplementedError(
                "Saving 'extended video' netCDFs is not supported for prediction-only "
                "data writers. Make sure to set `log_extended_video_netcdfs=False`."
            )
        return DataWriter(
            path=experiment_dir,
            n_initial_conditions=n_initial_conditions,
            n_timesteps=n_timesteps,
            variable_metadata=variable_metadata,
            coords=coords,
            timestep=timestep,
            enable_prediction_netcdfs=self.save_prediction_files,
            enable_monthly_netcdfs=self.save_monthly_files,
            save_names=self.names,
            time_coarsen=self.time_coarsen,
            dataset_metadata=dataset_metadata,
        )


class PairedDataWriter(WriterABC[PrognosticState, PairedData]):
    def __init__(
        self,
        path: str,
        n_initial_conditions: int,
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        timestep: datetime.timedelta,
        enable_prediction_netcdfs: bool,
        enable_monthly_netcdfs: bool,
        enable_video_netcdfs: bool,
        save_names: Sequence[str] | None,
        enable_histogram_netcdfs: bool,
        dataset_metadata: DatasetMetadata,
        time_coarsen: TimeCoarsenConfig | None = None,
    ):
        """
        Args:
            path: Path to write netCDF file(s).
            n_initial_conditions: Number of ICs/ensemble members to write to the file.
            n_timesteps: Number of timesteps to write to the file.
            variable_metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
            timestep: Timestep of the model.
            enable_prediction_netcdfs: Whether to enable writing of netCDF files
                containing the predictions and target values.
            enable_monthly_netcdfs: Whether to enable writing of netCDF files
                containing the monthly predictions and target values.
            enable_video_netcdfs: Whether to enable writing of netCDF files
                containing video metrics.
            save_names: Names of variables to save in the prediction, histogram,
                and monthly netCDF files.
            enable_histogram_netcdfs: Whether to write netCDFs with histogram data.
            dataset_metadata: Metadata for the dataset.
            time_coarsen: Configuration for time coarsening of written outputs.
        """
        self._writers: list[PairedSubwriter] = []
        self.path = path
        self.coords = coords
        self.variable_metadata = variable_metadata
        self.dataset_metadata = dataset_metadata

        if time_coarsen is not None:
            n_coarsened_timesteps = time_coarsen.n_coarsened_timesteps(n_timesteps)
        else:
            n_coarsened_timesteps = n_timesteps

        def _time_coarsen_builder(data_writer: PairedSubwriter) -> PairedSubwriter:
            if time_coarsen is not None:
                return time_coarsen.build_paired(data_writer)
            return data_writer

        if enable_prediction_netcdfs:
            self._writers.append(
                _time_coarsen_builder(
                    PairedRawDataWriter(
                        path=path,
                        n_initial_conditions=n_initial_conditions,
                        save_names=save_names,
                        variable_metadata=variable_metadata,
                        coords=coords,
                        dataset_metadata=dataset_metadata,
                    )
                )
            )
        if enable_monthly_netcdfs:
            self._writers.append(
                PairedMonthlyDataWriter(
                    path=path,
                    n_samples=n_initial_conditions,
                    n_timesteps=n_timesteps,
                    timestep=timestep,
                    save_names=save_names,
                    variable_metadata=variable_metadata,
                    coords=coords,
                    dataset_metadata=dataset_metadata,
                )
            )
        if enable_video_netcdfs:
            self._writers.append(
                _time_coarsen_builder(
                    PairedVideoDataWriter(
                        path=path,
                        n_timesteps=n_coarsened_timesteps,
                        variable_metadata=variable_metadata,
                        coords=coords,
                        dataset_metadata=dataset_metadata,
                    )
                )
            )
        if enable_histogram_netcdfs:
            self._writers.append(
                _time_coarsen_builder(
                    PairedHistogramDataWriter(
                        path=path,
                        n_timesteps=n_coarsened_timesteps,
                        variable_metadata=variable_metadata,
                        save_names=save_names,
                        dataset_metadata=dataset_metadata,
                    )
                )
            )
        self._n_timesteps_seen = 0

    def write(self, data: PrognosticState, filename: str):
        """Eagerly write data to a single netCDF file.

        Args:
            data: the data to be written.
            filename: the filename to use for the netCDF file.
        """
        _write(
            data=data.as_batch_data(),
            path=self.path,
            filename=filename,
            variable_metadata=self.variable_metadata,
            coords=self.coords,
            dataset_metadata=self.dataset_metadata,
        )

    def append_batch(
        self,
        batch: PairedData,
    ):
        """
        Append a batch of data to the file.

        Args:
            batch: Prediction and target data.
        """
        for writer in self._writers:
            writer.append_batch(
                target=dict(batch.reference),
                prediction=dict(batch.prediction),
                start_timestep=self._n_timesteps_seen,
                batch_time=batch.time,
            )
        self._n_timesteps_seen += batch.time.shape[1]

    def flush(self):
        """
        Flush the data to disk.
        """
        for writer in self._writers:
            writer.flush()


def _write(
    data: BatchData,
    path: str,
    filename: str,
    variable_metadata: Mapping[str, VariableMetadata],
    coords: Mapping[str, np.ndarray],
    dataset_metadata: DatasetMetadata,
):
    """Write provided data to a single netCDF at specified path/filename.

    If the data has only one timestep, the data is squeezed to remove
    the time dimension.

    Args:
        data: Batch data to written.
        path: Directory to write the netCDF file in.
        filename: filename to use for netCDF.
        variable_metadata: Metadata for each variable to be written to the file.
        coords: Coordinate data to be written to the file.
        dataset_metadata: Metadata for the dataset.
    """
    if data.time.sizes["time"] == 1:
        time_dim = data.dims.index("time")
        dims_to_write = data.dims[:time_dim] + data.dims[time_dim + 1 :]

        def maybe_squeeze(x: torch.Tensor) -> torch.Tensor:
            return x.squeeze(dim=time_dim)

        time_array = data.time.isel(time=0)
    else:
        dims_to_write = data.dims

        def maybe_squeeze(x):
            return x

        time_array = data.time

    data_arrays = {}
    for name in data.data:
        array = maybe_squeeze(data.data[name]).cpu().numpy()
        data_arrays[name] = xr.DataArray(array, dims=dims_to_write)
        if name in variable_metadata:
            data_arrays[name].attrs = {
                "long_name": variable_metadata[name].long_name,
                "units": variable_metadata[name].units,
            }
    data_arrays["time"] = time_array
    ds = xr.Dataset(data_arrays, coords=coords)
    ds.attrs.update(dataset_metadata.as_flat_str_dict())
    ds.to_netcdf(str(Path(path) / filename))


class DataWriter(WriterABC[PrognosticState, PairedData]):
    def __init__(
        self,
        path: str,
        n_initial_conditions: int,
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        timestep: datetime.timedelta,
        enable_prediction_netcdfs: bool,
        enable_monthly_netcdfs: bool,
        save_names: Sequence[str] | None,
        dataset_metadata: DatasetMetadata,
        time_coarsen: TimeCoarsenConfig | None = None,
    ):
        """
        Args:
            path: Directory within which to write netCDF file(s).
            n_initial_conditions: Number of initial conditions / timeseries
                to write to the file.
            n_timesteps: Number of timesteps to write to the file.
            variable_metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
            timestep: Timestep of the model.
            enable_prediction_netcdfs: Whether to enable writing of netCDF files
                containing the predictions and target values.
            enable_monthly_netcdfs: Whether to enable writing of netCDF files
            save_names: Names of variables to save in the prediction, histogram,
                and monthly netCDF files.
            dataset_metadata: Metadata for the dataset.
            time_coarsen: Configuration for time coarsening of raw outputs.
        """
        self._writers: list[Subwriter] = []
        if "face" in coords:
            # TODO: handle writing HEALPix data
            # https://github.com/ai2cm/full-model/issues/1089
            return

        def _time_coarsen_builder(data_writer: Subwriter) -> Subwriter:
            if time_coarsen is not None:
                return time_coarsen.build(data_writer)
            return data_writer

        if enable_prediction_netcdfs:
            self._writers.append(
                _time_coarsen_builder(
                    RawDataWriter(
                        path=path,
                        label="autoregressive_predictions.nc",
                        n_initial_conditions=n_initial_conditions,
                        save_names=save_names,
                        variable_metadata=variable_metadata,
                        coords=coords,
                        dataset_metadata=dataset_metadata,
                    )
                )
            )

        if enable_monthly_netcdfs:
            self._writers.append(
                MonthlyDataWriter(
                    path=path,
                    label="predictions",
                    n_samples=n_initial_conditions,
                    n_months=months_for_timesteps(n_timesteps, timestep),
                    save_names=save_names,
                    variable_metadata=variable_metadata,
                    coords=coords,
                    dataset_metadata=dataset_metadata,
                )
            )

        self.path = path
        self.variable_metadata = variable_metadata
        self.dataset_metadata = dataset_metadata
        self.coords = coords
        self._n_timesteps_seen = 0

    def append_batch(self, batch: PairedData):
        """
        Append prediction data to the file. The prognostic data
        and forcing data are merged before writing.

        Args:
            batch: Paired data to be written.
        """
        merged = {**batch.prediction, **batch.forcing}
        unpaired_batch = BatchData.new_on_device(
            data=merged,
            time=batch.time,
        )
        self._append_batch(unpaired_batch)

    def _append_batch(self, batch: BatchData):
        for writer in self._writers:
            writer.append_batch(
                data=dict(batch.data),
                start_timestep=self._n_timesteps_seen,
                batch_time=batch.time,
            )
        self._n_timesteps_seen += batch.time.shape[1]

    def flush(self):
        """
        Flush the data to disk.
        """
        for writer in self._writers:
            writer.flush()

    def write(self, data: PrognosticState, filename: str):
        _write(
            data=data.as_batch_data(),
            path=self.path,
            filename=filename,
            variable_metadata=self.variable_metadata,
            coords=self.coords,
            dataset_metadata=self.dataset_metadata,
        )



import torch

from fme.core.typing_ import TensorDict, TensorMapping


def force_positive(data: TensorMapping, names: list[str]) -> TensorDict:
    """Clamp all tensors defined by `names` to be greater than or equal to zero."""
    out = {**data}
    for name in names:
        out[name] = torch.clamp(data[name], min=0.0)
    return out



import datetime

import pytest
import torch

from fme.core.constants import DENSITY_OF_WATER_CM4, SPECIFIC_HEAT_OF_WATER_CM4
from fme.core.coordinates import DepthCoordinate
from fme.core.ocean_data import OceanData
from fme.core.ocean_derived_variables import (
    _compute_ocean_derived_variable,
    compute_ocean_derived_quantities,
    get_ocean_derived_variable_metadata,
)
from fme.core.typing_ import TensorDict, TensorMapping

TIMESTEP = datetime.timedelta(hours=5 * 24)


def test_compute_ocean_derived_variable():
    """Test computing a single ocean derived variable."""
    fake_data = {
        "thetao_0": torch.tensor([29.0]),
        "thetao_1": torch.tensor([10.0]),
    }

    idepth = torch.tensor([2.5, 10, 20])
    lev_thickness = idepth.diff(dim=-1)
    depth_coordinate = DepthCoordinate(
        idepth=idepth,
        mask=torch.ones(2),
    )

    def _derived_variable_func(data: OceanData, *_) -> torch.Tensor:
        return data.ocean_heat_content

    output_data = _compute_ocean_derived_variable(
        fake_data,
        depth_coordinate,
        TIMESTEP,
        "ocean_heat_content",
        _derived_variable_func,
    )
    assert "ocean_heat_content" in output_data
    torch.testing.assert_close(
        output_data["ocean_heat_content"],
        torch.tensor(
            [
                SPECIFIC_HEAT_OF_WATER_CM4
                * DENSITY_OF_WATER_CM4
                * (
                    lev_thickness[0] * fake_data["thetao_0"]
                    + lev_thickness[1] * fake_data["thetao_1"]
                )
            ]
        ),
    )


def test_compute_ocean_derived_variable_raises_value_error_when_overwriting():
    """Test that attempting to overwrite an existing variable raises an error."""
    fake_data = {
        "thetao_0": torch.tensor([29.0]),
        "thetao_1": torch.tensor([29.0]),
    }
    depth_coordinate = DepthCoordinate(
        idepth=torch.tensor([0.0, 5.0, 15.0]),
        mask=torch.ones(2),
    )

    def compute_ohc(data: OceanData, *_) -> torch.Tensor:
        return data.ocean_heat_content

    with pytest.raises(ValueError, match="already exists"):
        _compute_ocean_derived_variable(
            fake_data, depth_coordinate, TIMESTEP, "thetao_0", compute_ohc
        )


def test_compute_ocean_derived_variable_existing_variable():
    """Test that attempting to overwrite an existing variable raises an error."""
    fake_data = {
        "sea_ice_fraction": torch.tensor([1.0]),
    }
    depth_coordinate = DepthCoordinate(
        idepth=torch.tensor([0.0, 5.0, 15.0]),
        mask=torch.ones(2),
    )

    def modify_sea_ice_fraction(data: OceanData, *_) -> torch.Tensor:
        return data.sea_ice_fraction - 1

    new_data = _compute_ocean_derived_variable(
        fake_data,
        depth_coordinate,
        TIMESTEP,
        "sea_ice_fraction",
        modify_sea_ice_fraction,
        exists_ok=True,
    )
    torch.testing.assert_close(
        fake_data["sea_ice_fraction"],
        new_data["sea_ice_fraction"],
        msg=(
            "Existing variables should not be modified by "
            "_compute_ocean_derived_variable"
        ),
    )


def test_compute_ocean_derived_quantities():
    """Test computing all registered ocean derived variables."""
    torch.manual_seed(0)

    fake_data = {
        "thetao_0": torch.rand(2, 3, 4, 8),  # [batch, time, lat, lon]
        "thetao_1": torch.rand(2, 3, 4, 8),
        "ocean_sea_ice_fraction": torch.rand(2, 3, 4, 8),
        "land_fraction": torch.rand(2, 3, 4, 8),
    }
    gen_data = fake_data.copy()
    depth_coordinate = DepthCoordinate(
        idepth=torch.tensor([0.0, 5.0, 15.0]),
        mask=torch.ones(2, 3, 4, 8, 2),
    )

    def derive_func(data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        updated = compute_ocean_derived_quantities(
            dict(data),
            depth_coordinate=depth_coordinate,
            timestep=TIMESTEP,
            forcing_data=dict(forcing_data),
        )
        return updated

    out_data = derive_func(gen_data, fake_data)

    # Test that ocean_heat_content was computed
    assert "ocean_heat_content" in out_data
    assert out_data["ocean_heat_content"].shape == (2, 3, 4, 8)

    # Test that sea_ice_fraction was computed
    assert "sea_ice_fraction" in out_data
    assert out_data["sea_ice_fraction"].shape == (2, 3, 4, 8)


def test_metadata_registry():
    """Test that the metadata registry contains expected entries."""
    metadata = get_ocean_derived_variable_metadata()
    assert metadata["ocean_heat_content"].units == "J/m**2"
    assert (
        metadata["ocean_heat_content"].long_name
        == "Column-integrated ocean heat content"
    )



import dataclasses
import datetime
from collections.abc import Callable, Mapping
from typing import Any, Literal, Protocol

import dacite
import torch

import fme
from fme.core.atmosphere_data import (
    AtmosphereData,
    HasAtmosphereVerticalIntegral,
    compute_layer_thickness,
)
from fme.core.constants import GRAVITY, SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME
from fme.core.corrector.registry import CorrectorABC
from fme.core.corrector.utils import force_positive
from fme.core.gridded_ops import GriddedOperations
from fme.core.registry.corrector import CorrectorSelector
from fme.core.typing_ import TensorDict, TensorMapping


@dataclasses.dataclass
class EnergyBudgetConfig:
    """Configuration for total energy budget correction.

    Parameters:
        method: Method to use for total energy budget correction. The available option
            is "constant_temperature", which enforces conservation of total energy by
            imposing a vertically and horizontally uniform air temperature correction.
        constant_unaccounted_heating: Column-integrated heating in W/m**2 to be added
            to the energy flux into the atmosphere when conserving total energy.
            This can be useful for correcting errors in energy budget in target data.
            The same additional heating is imposed at all time steps and grid cells.
    """

    method: Literal["constant_temperature"]
    constant_unaccounted_heating: float = 0.0


@CorrectorSelector.register("atmosphere_corrector")
@dataclasses.dataclass
class AtmosphereCorrectorConfig:
    r"""
    Configuration for the post-step state corrector.

    ``conserve_dry_air`` enforces the constraint that:

    .. math::

        global\_dry\_air = global\_mean(ps -
        sum_k((ak\_diff + bk\_diff \* ps) \* wat_k))

    in the generated data is equal to its value in the input data. This is done
    by adding a globally-constant correction to the surface pressure in each
    column. As per-mass values such as mixing ratios of water are unchanged,
    this can cause changes in total water or energy. Note all global means here
    are area-weighted.

    ``zero_global_mean_moisture_advection`` enforces the constraint that:

    .. math::

        global\_mean(tendency\_of\_total\_water\_path\_due\_to\_advection) = 0

    in the generated data. This is done by adding a globally-constant correction
    to the moisture advection tendency in each column.

    ``moisture_budget_correction`` enforces closure of the moisture budget equation:

    .. math::

        tendency\_of\_total\_water\_path = (evaporation\_rate - precipitation\_rate
        \\\\ + tendency\_of\_total\_water\_path\_due\_to\_advection)

    in the generated data, where ``tendency_of_total_water_path`` is the difference
    between the total water path at the current timestep and the previous
    timestep divided by the time difference. This is done by modifying the
    precipitation, evaporation, and/or moisture advection tendency fields as
    described in the ``moisture_budget_correction`` attribute. When
    advection tendency is modified, this budget equation is enforced in each
    column, while when only precipitation or evaporation are modified, only
    the global mean of the budget equation is enforced.

    When enforcing moisture budget closure, we assume the global mean moisture
    advection is zero. Therefore ``zero_global_mean_moisture_advection`` must be
    True if using a ``moisture_budget_correction`` option other than ``None``.

    Parameters:
        conserve_dry_air: If True, force the generated data to conserve dry air
            by subtracting a constant offset from the surface pressure of each
            column. This can cause changes in per-mass values such as total water
            or energy.
        zero_global_mean_moisture_advection: If True, force the generated data to
            have zero global mean moisture advection by subtracting a constant
            offset from the moisture advection tendency of each column.
        moisture_budget_correction: If not "None", force the generated data to
            conserve global or column-local moisture by modifying budget fields.
            Options are:

            - ``precipitation``: multiply precipitation by a scale factor
              to close the global moisture budget.
            - ``evaporation``: multiply evaporation by a scale factor
              to close the global moisture budget.
            - ``advection_and_precipitation``: after applying the "precipitation"
              global-mean correction above, recompute the column-integrated
              advective tendency as the budget residual,
              ensuring column budget closure.
            - ``advection_and_evaporation``: after applying the "evaporation"
              global-mean correction above, recompute the column-integrated
              advective tendency as the budget residual,
              ensuring column budget closure.

        force_positive_names: Names of fields that should be forced to be greater
            than or equal to zero. This is useful for fields like precipitation.
        total_energy_budget_correction: If not None, force the generated data to
            conserve an idealized version of total energy using the provided
            configuration.
    """

    conserve_dry_air: bool = False
    zero_global_mean_moisture_advection: bool = False
    moisture_budget_correction: (
        Literal[
            "precipitation",
            "evaporation",
            "advection_and_precipitation",
            "advection_and_evaporation",
        ]
        | None
    ) = None
    force_positive_names: list[str] = dataclasses.field(default_factory=list)
    total_energy_budget_correction: EnergyBudgetConfig | None = None

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "AtmosphereCorrectorConfig":
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )


class AtmosphereCorrector(CorrectorABC):
    def __init__(
        self,
        config: AtmosphereCorrectorConfig,
        gridded_operations: GriddedOperations,
        vertical_coordinate: HasAtmosphereVerticalIntegral | None,
        timestep: datetime.timedelta,
    ):
        self._config = config
        self._gridded_operations = gridded_operations
        self._vertical_coordinate = vertical_coordinate

        self._timestep_seconds = timestep.total_seconds()
        if fme.get_device() == torch.device("mps", 0):
            self._dry_air_precision = torch.float32
        else:
            self._dry_air_precision = torch.float64

    def __call__(
        self,
        input_data: TensorMapping,
        gen_data: TensorMapping,
        forcing_data: TensorMapping,
    ) -> TensorDict:
        """Apply corrections to the generated data.

        Args:
            input_data: The input time step data.
            gen_data: The data generated by the model, to be corrected.
            forcing_data: The forcing data for the same time step as gen_data.

        Returns:
            The corrected data.
        """
        gen_data = dict(gen_data)
        if len(self._config.force_positive_names) > 0:
            # do this step before imposing other conservation correctors, since
            # otherwise it could end up creating violations of those constraints.
            gen_data = force_positive(gen_data, self._config.force_positive_names)
        if self._config.conserve_dry_air:
            if self._vertical_coordinate is None:
                raise ValueError(
                    "conserve_dry_air is set to True, but no vertical coordinate is "
                    "available."
                )
            gen_data = _force_conserve_dry_air(
                input_data=input_data,
                gen_data=gen_data,
                area_weighted_mean=self._gridded_operations.area_weighted_mean,
                vertical_coordinate=self._vertical_coordinate,
                precision=self._dry_air_precision,
            )
        if self._config.zero_global_mean_moisture_advection:
            gen_data = _force_zero_global_mean_moisture_advection(
                gen_data=gen_data,
                area_weighted_mean=self._gridded_operations.area_weighted_mean,
            )
        if self._config.moisture_budget_correction is not None:
            if self._vertical_coordinate is None:
                raise ValueError(
                    "Moisture budget correction is turned on, but no vertical "
                    "coordinate is available."
                )
            gen_data = _force_conserve_moisture(
                input_data=input_data,
                gen_data=gen_data,
                area_weighted_mean=self._gridded_operations.area_weighted_mean,
                vertical_coordinate=self._vertical_coordinate,
                timestep_seconds=self._timestep_seconds,
                terms_to_modify=self._config.moisture_budget_correction,
            )
        if self._config.total_energy_budget_correction is not None:
            if self._vertical_coordinate is None:
                raise ValueError(
                    "Energy budget correction is turned on, but no vertical coordinate"
                    " is available."
                )
            gen_data = _force_conserve_total_energy(
                input_data=input_data,
                gen_data=gen_data,
                forcing_data=forcing_data,
                area_weighted_mean=self._gridded_operations.area_weighted_mean,
                vertical_coordinate=self._vertical_coordinate,
                timestep_seconds=self._timestep_seconds,
                method=self._config.total_energy_budget_correction.method,
                unaccounted_heating=self._config.total_energy_budget_correction.constant_unaccounted_heating,
            )
        return gen_data


class AreaWeightedMean(Protocol):
    def __call__(
        self, data: torch.Tensor, keepdim: bool, name: str | None = None
    ) -> torch.Tensor: ...


def _force_conserve_dry_air(
    input_data: TensorMapping,
    gen_data: TensorMapping,
    area_weighted_mean: AreaWeightedMean,
    vertical_coordinate: HasAtmosphereVerticalIntegral,
    precision: torch.dtype = torch.float64,
) -> TensorDict:
    """
    Update the generated data to conserve dry air.

    This is done by adding a constant correction to the dry air pressure of
    each column, and may result in changes in per-mass values such as
    total water or energy.

    We first compute the target dry air pressure by computing the globally
    averaged difference in dry air pressure between the input_data and gen_data,
    and then add this offset to the fully-resolved gen_data dry air pressure.
    We can then solve for the surface pressure corresponding to this new dry air
    pressure.

    We start from the expression for dry air pressure:

        dry_air = ps - sum_k((ak_diff + bk_diff * ps) * wat_k)

    To update the dry air, we compute and update the surface pressure:

        ps = (
            dry_air + sum_k(ak_diff * wat_k)
        ) / (
            1 - sum_k(bk_diff * wat_k)
        )
    """
    input = AtmosphereData(input_data, vertical_coordinate)
    if input.surface_pressure is None:
        raise ValueError("surface_pressure is required to force dry air conservation")
    gen = AtmosphereData(gen_data, vertical_coordinate)
    gen_dry_air = gen.surface_pressure_due_to_dry_air
    global_gen_dry_air = area_weighted_mean(gen_dry_air.to(precision), keepdim=True)
    global_target_gen_dry_air = area_weighted_mean(
        input.surface_pressure_due_to_dry_air.to(precision),
        keepdim=True,
    )
    error = global_gen_dry_air - global_target_gen_dry_air
    new_gen_dry_air = gen_dry_air.to(precision) - error
    try:
        wat = gen.specific_total_water.to(precision)
    except KeyError:
        raise ValueError("specific_total_water is required for conservation")
    ak_diff = vertical_coordinate.get_ak().diff().to(precision)
    bk_diff = vertical_coordinate.get_bk().diff().to(precision)
    new_pressure = (new_gen_dry_air + (ak_diff * wat).sum(-1)) / (
        1 - (bk_diff * wat).sum(-1)
    )
    gen.set_surface_pressure(new_pressure.to(dtype=input.surface_pressure.dtype))
    return gen.data


def _force_zero_global_mean_moisture_advection(
    gen_data: TensorMapping,
    area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
) -> TensorDict:
    """
    Update the generated data so advection conserves moisture.

    Does so by adding a constant offset to the moisture advective tendency.

    Args:
        gen_data: The generated data.
        area_weighted_mean: Computes an area-weighted mean,
            removing horizontal dimensions.
    """
    gen = AtmosphereData(gen_data)

    mean_moisture_advection = area_weighted_mean(
        gen.tendency_of_total_water_path_due_to_advection,
    )
    gen.set_tendency_of_total_water_path_due_to_advection(
        gen.tendency_of_total_water_path_due_to_advection
        - mean_moisture_advection[..., None, None]
    )
    return gen.data


def _force_conserve_moisture(
    input_data: TensorMapping,
    gen_data: TensorMapping,
    area_weighted_mean: AreaWeightedMean,
    vertical_coordinate: HasAtmosphereVerticalIntegral,
    timestep_seconds: float,
    terms_to_modify: Literal[
        "precipitation",
        "evaporation",
        "advection_and_precipitation",
        "advection_and_evaporation",
    ],
) -> TensorDict:
    """
    Update the generated data to conserve moisture.

    Does so while conserving total dry air in each column.

    Assumes the global mean advective tendency of moisture is zero. This assumption
    means any existing global mean advective tendency will be set to zero
    if the advective tendency is re-computed.

    Args:
        input_data: The input data.
        gen_data: The generated data one timestep after the input data.
        area_weighted_mean: Computes an area-weighted mean,
            removing horizontal dimensions.
        vertical_coordinate: The sigma coordinates.
        timestep_seconds: Timestep of the model in seconds.
        terms_to_modify: Which terms to modify, in addition to modifying surface
            pressure to conserve dry air mass. One of:
            - "precipitation": modify precipitation only
            - "evaporation": modify evaporation only
            - "advection_and_precipitation": modify advection and precipitation
            - "advection_and_evaporation": modify advection and evaporation
    """
    input = AtmosphereData(input_data, vertical_coordinate)
    gen = AtmosphereData(gen_data, vertical_coordinate)

    gen_total_water_path = gen.total_water_path
    twp_total_tendency = (
        gen_total_water_path - input.total_water_path
    ) / timestep_seconds
    twp_tendency_global_mean = area_weighted_mean(twp_total_tendency, keepdim=True)
    evaporation_global_mean = area_weighted_mean(gen.evaporation_rate, keepdim=True)
    precipitation_global_mean = area_weighted_mean(gen.precipitation_rate, keepdim=True)
    if terms_to_modify.endswith("precipitation"):
        # We want to achieve
        #     global_mean(twp_total_tendency) = (
        #         global_mean(evaporation_rate)
        #         - global_mean(precipitation_rate)
        #     )
        # so we modify precipitation_rate to achieve this. Note we have
        # assumed the global mean advection tendency is zero.
        # First, we find the required global-mean precipitation rate
        #     new_global_precip_rate = (
        #         global_mean(evaporation_rate)
        #         - global_mean(twp_total_tendency)
        #     )
        new_precipitation_global_mean = (
            evaporation_global_mean - twp_tendency_global_mean
        )
        # Because scalar multiplication commutes with summation, we can
        # achieve this by multiplying each gridcell's precipitation rate
        # by the ratio of the new global mean to the current global mean.
        #    new_precip_rate = (
        #        new_global_precip_rate / current_global_precip_rate
        #    ) * current_precip_rate
        gen.set_precipitation_rate(
            gen.precipitation_rate
            * (new_precipitation_global_mean / precipitation_global_mean)
        )
    elif terms_to_modify.endswith("evaporation"):
        # Derived similarly as for "precipitation" case.
        new_evaporation_global_mean = (
            twp_tendency_global_mean + precipitation_global_mean
        )
        gen.set_evaporation_rate(
            gen.evaporation_rate
            * (new_evaporation_global_mean / evaporation_global_mean)
        )
    if terms_to_modify.startswith("advection"):
        # Having already corrected the global-mean budget, we recompute
        # advection based on assumption that the columnwise
        # moisture budget closes. Correcting the global mean budget first
        # is important to ensure the resulting advection has zero global mean.
        new_advection = twp_total_tendency - (
            gen.evaporation_rate - gen.precipitation_rate
        )
        gen.set_tendency_of_total_water_path_due_to_advection(new_advection)
    return gen.data


def _force_conserve_total_energy(
    input_data: TensorMapping,
    gen_data: TensorMapping,
    forcing_data: TensorMapping,
    area_weighted_mean: AreaWeightedMean,
    vertical_coordinate: HasAtmosphereVerticalIntegral,
    timestep_seconds: float,
    method: Literal["constant_temperature"] = "constant_temperature",
    unaccounted_heating: float = 0.0,
) -> TensorDict:
    """Apply a correction to the generated data to conserve total energy.

    This function also inserts the unaccounted heating into the generated data.
    """
    if method != "constant_temperature":
        raise NotImplementedError(
            f"Method {method} not implemented for total energy conservation"
        )
    input = AtmosphereData(input_data, vertical_coordinate)
    forcing = AtmosphereData(forcing_data)
    required_forcing = {
        "DSWRFtoa": forcing.toa_down_sw_radiative_flux,
        "HGTsfc": forcing.surface_height,
    }
    atmosphere_data = dict(gen_data)
    for name, tensor in required_forcing.items():
        atmosphere_data[name] = tensor
    gen = AtmosphereData(atmosphere_data, vertical_coordinate)

    gen_energy_path = gen.total_energy_ace2_path
    input_energy_path = input.total_energy_ace2_path
    predicted_energy_flux_into_atmosphere = gen.net_energy_flux_into_atmosphere

    gen_energy_path_global_mean = area_weighted_mean(gen_energy_path, keepdim=True)
    input_energy_path_global_mean = area_weighted_mean(input_energy_path, keepdim=True)
    energy_flux_global_mean = area_weighted_mean(
        predicted_energy_flux_into_atmosphere, keepdim=True
    )

    desired_energy_path_global_mean = (
        input_energy_path_global_mean
        + (energy_flux_global_mean + unaccounted_heating) * timestep_seconds
    )

    energy_correction = desired_energy_path_global_mean - gen_energy_path_global_mean
    energy_to_temperature_factor = _energy_correction_factor(gen, vertical_coordinate)
    # take global mean to impose a spatially uniform temperature correction
    energy_to_temp_factor_gm = area_weighted_mean(energy_to_temperature_factor, True)
    temperature_correction = energy_correction / energy_to_temp_factor_gm

    # apply same temperature correction to all vertical layers
    n_levels = gen.air_temperature.shape[-1]
    for k in range(n_levels):
        name = f"air_temperature_{k}"
        gen.data[name] = gen.data[name] + torch.nan_to_num(
            temperature_correction, nan=0.0
        )

    # filter required here because we merged forcing data into gen above
    return {k: v for k, v in gen.data.items() if k in gen_data}


def _energy_correction_factor(
    gen: AtmosphereData, vertical_coordinate: HasAtmosphereVerticalIntegral
) -> torch.Tensor:
    """
    Compute the factor to get a vertically-uniform temperature correction that
    will lead to a desired change in the globally-averaged total energy.

    See https://www.overleaf.com/read/dqjjcvzxnfvn#d525aa.
    """
    interface_pressure = vertical_coordinate.interface_pressure(gen.surface_pressure)
    q_times_dlogp = (
        compute_layer_thickness(
            interface_pressure, gen.air_temperature, gen.specific_total_water
        )
        * GRAVITY
        / gen.air_temperature
    )
    cumulative = torch.cumsum(q_times_dlogp.flip(dims=(-1,)), dim=-1).flip(dims=(-1,))
    total_integrand = (
        SPECIFIC_HEAT_OF_DRY_AIR_CONST_VOLUME - 0.5 * q_times_dlogp + cumulative
    )
    correction_factor = vertical_coordinate.vertical_integral(
        total_integrand, gen.surface_pressure
    )
    return correction_factor



import dataclasses
from collections.abc import Mapping
from typing import Any

import torch.optim.lr_scheduler


@dataclasses.dataclass
class SchedulerConfig:
    """
    Configuration for a scheduler to use during training.

    Parameters:
        type: Name of scheduler class from torch.optim.lr_scheduler,
            no scheduler is used by default.
        kwargs: Keyword arguments to pass to the scheduler constructor.
    """

    type: str | None = None
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)

    def build(
        self, optimizer, max_epochs
    ) -> torch.optim.lr_scheduler._LRScheduler | None:
        """
        Build the scheduler.
        """
        if self.type is None:
            return None

        build_kwargs = {**self.kwargs}
        # work-around so we don't need to specify T_max
        # in the yaml file for this scheduler
        if self.type == "CosineAnnealingLR" and "T_max" not in self.kwargs:
            build_kwargs["T_max"] = max_epochs

        scheduler_class = getattr(torch.optim.lr_scheduler, self.type)
        return scheduler_class(optimizer=optimizer, **build_kwargs)
